//************************************************************************
//*                          BipFILE Win V0.9.1                          *
//*----------------------------------------------------------------------*
//*                BipFILE FUNCTIONS for Microsoft Windows               *
//*                                                                      *
//*                                                                      *
//* Licence : NFM                                                        *
//*                                                         P. Basset    *
//************************************************************************


#pragma warning (disable:4786)	// disable warning "identifier was truncated to '255' characters" when using std::map


#include <windows.h>
#include <string>
#include <map>
#include <sstream>
#include <iomanip>
#include <time.h>

#include "BipFILE.h"
#include "BipDATE.h"


	using namespace std;


	// Global Variables : can be modify
	string BF_Language				= "FR";		// default language

	bool BF_DISPLAY_ERROR			= true;		// if true then display error message
	bool BF_WRITE_ERROR				= false;	// if true then error message is write in biperror.txt file

	BF_STRUCT_ERROR BFError;					// structure for BipFILE error
	BF_STRUCT_ERROR BFErrorSave;				// structure for BipFILE error


	// #### Global Variables #### : **** DON'T MOFIFY ****
	const string BF_PLATEFORM		= "WIN";		// plateform of BipFile
	const string BF_VERSION			= "0.9.1";		// version of BipFile

	// table declaration
	typedef map <string, BF_STRUCT_TABLE> BFTableDef;// map for table
	BFTableDef BFTable;								// declare table map
	typedef BFTableDef::iterator BFTableItDef;		// iterator for table map

	// field declaration
	typedef map <string, BF_STRUCT_FIELD> BFFieldDef;// map for field
	BFFieldDef BFField;								// declare field map
	typedef BFFieldDef::iterator BFFieldItDef;		// iterator for field map

	// stream declaration
	const int BF_MAX_STREAM			= 5;			// max number of stream
	BF_STRUCT_STREAM BFStream [BF_MAX_STREAM];		// structure for stream

	// #### End of Global Variables #### : **** DON'T MOFIFY ****





//-----------------------------------------------------------------
//!    BF_TableOpen : Open a database table                       !
//-----------------------------------------------------------------
//! sBFName : internal table name                                 !
//! sBFTableName : real table name in DataBase                    !
//! sBFDbName : OS DataBase name                                  !
//! sBFSrvName : server name                                      !
//! sBFDbType : DataBase type (ie : BDB = Bip DataBase)           !
//! sBFParam : param for opening                                  !
//!																  !
//! return : error or 0 if none                                   !
//-----------------------------------------------------------------
int BF_TableOpen (const string sBFName, const string sBFTableName, const string sBFDbName,
				  const string sBFSrvName, const string sBFDbType, const string sBFParam)
				  // sBFParam = ""
{
	int iError;
	int iStream;
	string sLine;
	string sDriver;
	string sMode;
	string sString;
	SQLRETURN SQLError;
	BF_STRUCT_TABLE TmpTable;
	bool bQuery;				// connection is for a query


	// for error management
	BF_ErrorSet (sBFName, "BF_TableOpen", "");


	// initialize variable
	iStream = -1;

	// internal table name is empty
	if (sBFName == "")
	{
		BF_ErrorDisplay (BF_ERR_NAME);
		return BF_ERR_NAME;
	}

	// verify if file name contain "." character
	iError = sBFName.find ('.', 0);
	if (iError >= 0)
	{
		BF_ErrorDisplay (BF_ERR_NOPOINT);
		return BF_ERR_NOPOINT;
	}

	// get query parameter
	sString = BF_StringParamGetIntl (BF_PRM_QUERY, sBFParam);
	if (BF_StringIsTrue (sString))
		bQuery = true;
	else
		bQuery = false;

	// verify table name
	if (sBFDbType != BF_DB_BDB && !bQuery)
	{
		// database table name is empty
		if (sBFTableName == "")
		{
			BF_ErrorDisplay (BF_ERR_BADDBTNAME);
			return BF_ERR_BADDBTNAME;
		}
	}

	// get table structure
	TmpTable = BF_TableGet (sBFName);
	if (TmpTable.sDbName != "")
	{
		BF_ErrorDisplay (BF_ERR_ALROPEN);
		return BF_ERR_ALROPEN;
	}

	// empty field structure
	TmpTable = BF_TableNullGet ();



	// BDB database
	if (sBFDbType == BF_DB_BDB)
	{
		// search first free stream
		for (iStream = 0; iStream < BF_MAX_STREAM; iStream++)
		{
			if (!BFStream [iStream].BFInUse)
				break;
		}

		// stream structure is full
		if (iStream >= BF_MAX_STREAM)
		{
			BF_ErrorDisplay (BF_ERR_STREAMFULL);
			return BF_ERR_STREAMFULL;
		}

		sMode = BF_StringParamGetIntl (BF_PRM_MODE, sBFParam);
		sMode = BF_StringToUpper (sMode);

		// open stream file
		if (sMode == "INPUT" || sMode == "")
		{
			BFStream [iStream].BFStream.open (sBFDbName.data(), ios::in);
			sMode = "INPUT";
		}
		else if (sMode == "APPEND")
			BFStream [iStream].BFStream.open (sBFDbName.data(), ios::out | ios::app);
		else if (sMode == "INOUT")
			BFStream [iStream].BFStream.open (sBFDbName.data(), ios::in | ios::out);
		else if (sMode == "OUTPUT")
			BFStream [iStream].BFStream.open (sBFDbName.data(), ios::out);
		else
		{
			BF_ErrorDisplay (BF_ERR_BADMODE);
			return BF_ERR_BADMODE;
		}

		BFStream [iStream].BFMode = sMode;


		// error during opening file
		if (BFStream [iStream].BFStream.fail ())
		{
			BFStream [iStream].BFStream.clear ();
			return BF_ERR_OPEN;
		}

		BFStream [iStream].BFInUse = true;
		BFStream [iStream].BFStream.clear ();
	}
	// SQL database
	else if (sBFDbType == BF_DB_ACCESS || sBFDbType == BF_DB_SQLSERVER ||
			 sBFDbType == BF_DB_MYSQL)
	{
		// environment handle
		SQLError = SQLAllocHandle (SQL_HANDLE_ENV, SQL_NULL_HANDLE,
									&TmpTable.SQLEnvir);    
		if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
		{
			BF_ErrorDisplay (BF_ERR_SQLALLOCENV);
			return BF_ERR_SQLALLOCENV;
		}

		// environment attribute
		SQLError = SQLSetEnvAttr (TmpTable.SQLEnvir, SQL_ATTR_ODBC_VERSION,
								  (SQLPOINTER)SQL_OV_ODBC3, SQL_IS_INTEGER);
		if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
		{
			BF_ErrorDisplay (BF_ERR_SQLSETENVATTR);
			return BF_ERR_SQLSETENVATTR;
		}

		// connection handle
		SQLError = SQLAllocHandle (SQL_HANDLE_DBC, TmpTable.SQLEnvir, &TmpTable.SQLDbc);
		if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
		{
			BF_ErrorDisplay (BF_ERR_SQLSETENVATTR);
			return BF_ERR_SQLSETENVATTR;
		}

		// connection attribute
		SQLError = SQLSetConnectAttr (TmpTable.SQLDbc, SQL_ATTR_ACCESS_MODE,
									   (SQLPOINTER) SQL_MODE_READ_WRITE, 0);
		if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
		{
			BF_ErrorDisplay (BF_ERR_SQLSETCONNATTR);
			return BF_ERR_SQLSETCONNATTR;
		}


		// prepare connection string
		sDriver = BF_SQLConnectStringGet (sBFDbName, sBFSrvName, sBFDbType, sBFParam);

		// connection to the driver
		SQLError = SQLDriverConnect (TmpTable.SQLDbc, NULL,
							(SQLCHAR*) sDriver.c_str(), sDriver.length (),
							0, 0, 0, SQL_DRIVER_COMPLETE_REQUIRED);
		if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
		{
			BF_ErrorDisplay (BF_ERR_SQLDRIVERCONNECT);
			return BF_ERR_SQLDRIVERCONNECT;
		}
	}
	else
	{
		// DataBase Type is unknown
		BF_ErrorDisplay (BF_ERR_BADTYPE);
		return BF_ERR_BADTYPE;
	}


	// DataBase is open
	TmpTable.sDbName = sBFDbName;
	TmpTable.sDbTableName = sBFTableName;
	TmpTable.sDbType = sBFDbType;
	TmpTable.sDbServer = sBFSrvName;
	TmpTable.iStream = iStream;

	// save current structure
	BFTable [sBFName] = TmpTable;

	// if BdType is "BDB" then load the file structure
	if (sBFDbType == BF_DB_BDB && sMode == "INPUT")
	{
		// load record description
		iError = BF_BDBRecordDescriptionLoad (sBFName);
		if (iError)
		{
// close file ?
			BF_ErrorDisplay (BF_ERR_OPEN);
			return BF_ERR_OPEN;
		}
	}
	// load SQL structure
	else if (sBFDbType != BF_DB_BDB && !bQuery)
	{
		iError = BF_SQLRecordDescriptionLoad (sBFName);
		if (iError)
		{
			// close file
			iError = BF_TableClose (sBFName);

			BF_ErrorDisplay (BF_ERR_OPEN);
			return BF_ERR_OPEN;
		}
	}


	return 0;

}



//-----------------------------------------------------------------
//!    BF_SQLRecordDescriptionLoad : load fields description      !
//-----------------------------------------------------------------
//! sBFName : internal table name                                 !
//!																  !
//! return : error or 0 if none                                   !
//-----------------------------------------------------------------
//! call by BF_OpenTable                                          !
//-----------------------------------------------------------------
int BF_SQLRecordDescriptionLoad (const string sBFName)
{
	int iError;
	SQLRETURN SQLError;
	SQLINTEGER iSQLLength;
	char szColName [BF_MAX_FIELDSIZE];
	char szColType [BF_MAX_FIELDSIZE];
	char szColRemark [BF_MAX_FIELDSIZE];
	char szDefValue [BF_MAX_FIELDSIZE];
	string sColType;
	SQLCHAR SQLChar [BF_MAX_FIELDSIZE];
	long lColSize;
	short int iColNull;
	bool bColNull;
	short int iColDate;
	short int iColDecimal;
	BF_STRUCT_TABLE TmpTable;
	SQLHSTMT hstmt;
	short int iColDateInt;
	short int iColDateSub;


	// for error management
	BF_ErrorSet (sBFName, "BF_SQLRecordDescriptionLoad", "");
	

	// read table structure
	TmpTable = BF_TableGet (sBFName);
	if (TmpTable.sDbName == "")
	{
		BF_ErrorDisplay (BF_ERR_NOTFOUND);
		return BF_ERR_NOTFOUND;
	}

	// verify type
	if (TmpTable.sDbType == BF_DB_BDB)
	{
		// DataBase Type is bad
		BF_ErrorDisplay (BF_ERR_BADTYPE);
		return BF_ERR_BADTYPE;
	}


	// declare statement handle
	SQLError = SQLAllocHandle (SQL_HANDLE_STMT, TmpTable.SQLDbc, &hstmt);
	if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
	{
		BF_ErrorDisplay (BF_ERR_SQLALLOCSTATEMENT);
		return BF_ERR_SQLALLOCSTATEMENT;
	}

	// reset internal statement handle
	TmpTable.SQLStmt = hstmt;

	// save current structure
	BFTable [sBFName] = TmpTable;

	// select all fields from the current table
	strncpy ( (char*) SQLChar, TmpTable.sDbTableName.c_str(),
		     TmpTable.sDbTableName.length() + 1);
	SQLError = SQLColumns (hstmt, NULL, 0, NULL, 0,
						   SQLChar, TmpTable.sDbTableName.length(), NULL, 0);
	if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
	{
			BF_ErrorDisplay (BF_ERR_SQLCOLUMNS);
			return BF_ERR_SQLCOLUMNS;
	}

	// read next line
	SQLError = SQLFetchScroll (hstmt, SQL_FETCH_NEXT, 0);
	if (SQLError == SQL_NO_DATA)
	{
		return -1;
	}
	if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
	{
		BF_ErrorDisplay (BF_ERR_SQLFETCHSCROLL);
		return BF_ERR_SQLFETCHSCROLL;
	}

	// get fields info
	while (SQLError == 0)
	{
		// get column name
		SQLError = SQLGetData (hstmt, 4, SQL_C_CHAR,
							   szColName, sizeof (szColName), &iSQLLength);
		if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
		{
			BF_ErrorSet ("", "", "Column 4");
			BF_ErrorDisplay (BF_ERR_SQLGETDATA);
			return BF_ERR_SQLGETDATA;
		}

		// get detail datetime type
		SQLError = SQLGetData (hstmt, 5, SQL_C_SSHORT,
							   &iColDate, 0, &iSQLLength);
		if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
		{
			BF_ErrorSet ("", "", "Column 5");
			BF_ErrorDisplay (BF_ERR_SQLGETDATA);
			return BF_ERR_SQLGETDATA;
		}

		// initialise number of character after decimal point
		iColDecimal = 0;

		// get column type
		SQLError = SQLGetData (hstmt, 6, SQL_C_CHAR,
							   szColType, sizeof (szColType), &iSQLLength);
		if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
		{
			BF_ErrorSet ("", "", "Column 6");
			BF_ErrorDisplay (BF_ERR_SQLGETDATA);
			return BF_ERR_SQLGETDATA;
		}

		// transform col type
		sColType = szColType;
		sColType = BF_StringToUpper (sColType);
		if (sColType == "VARCHAR" || sColType == "CHAR" || sColType == "LONGCHAR")
		{
			sColType = BF_FLD_STRING;
		}
		else if (sColType == "BYTE")
		{
			sColType = BF_FLD_BYTE;
		}
		else if (sColType == "SMALLINT" || sColType == "INT")
		{
			sColType = BF_FLD_INTEGER;
		}
		else if (sColType == "INTEGER")
		{
			sColType = BF_FLD_LONG;
		}
		else if (sColType == "REAL")
		{
			sColType = BF_FLD_FLOAT;
		}
		else if (sColType == "DATETIME")
		{
			if (iColDate == SQL_TYPE_TIMESTAMP)
			{
				sColType = BF_FLD_DATETIME;
			}
			else
			{
				BF_ErrorSet ("", "", "Bad DATETIME field : " + BF_IntToString(iColDate));
				BF_ErrorDisplay (BF_ERR_BADFIELDTYPE);
				return BF_ERR_BADFIELDTYPE;
			}

		}
		else if (sColType == "CURRENCY")
		{
			sColType = BF_FLD_CURRENCY;

			// get detail datetime type
			SQLError = SQLGetData (hstmt, 9, SQL_C_SSHORT,
								   &iColDecimal, 0, &iSQLLength);
			if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
			{
				BF_ErrorSet ("", "", "Column 9");
				BF_ErrorDisplay (BF_ERR_SQLGETDATA);
				return BF_ERR_SQLGETDATA;
			}

		}
		else
		{
			BF_ErrorSet ("", "", "Bad type field : " + sColType);
			BF_ErrorDisplay (BF_ERR_BADFIELDTYPE);
			return BF_ERR_BADFIELDTYPE;
		}
		
		// get column length
		SQLError = SQLGetData (hstmt, 7, SQL_C_SLONG,
							   &lColSize, 0, &iSQLLength);
		if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
		{
			BF_ErrorSet ("", "", "Column 7");
			BF_ErrorDisplay (BF_ERR_SQLGETDATA);
			return BF_ERR_SQLGETDATA;
		}

		// get nullable state
		SQLError = SQLGetData (hstmt, 11, SQL_C_SSHORT,
							   &iColNull, 0, &iSQLLength);
		if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
		{
			BF_ErrorSet ("", "", "Column 11");
			BF_ErrorDisplay (BF_ERR_SQLGETDATA);
			return BF_ERR_SQLGETDATA;
		}

		// transform null state
		bColNull = false;
		if (iColNull == SQL_NULLABLE)
			bColNull = true;
		
		
		// get remark
		SQLError = SQLGetData (hstmt, 12, SQL_C_CHAR,
							   &szColRemark, sizeof (szColRemark), &iSQLLength);
		if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
		{
			BF_ErrorSet ("", "", "Column 12");
			BF_ErrorDisplay (BF_ERR_SQLGETDATA);
			return BF_ERR_SQLGETDATA;
		}
		// remark is null
		if (iSQLLength == -1)
			szColRemark [0] = '\0';

		// get default value
		SQLError = SQLGetData (hstmt, 13, SQL_C_CHAR,
							   &szDefValue, sizeof (szDefValue), &iSQLLength);
		if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
		{
			BF_ErrorSet ("", "", "Column 13");
			BF_ErrorDisplay (BF_ERR_SQLGETDATA);
			return BF_ERR_SQLGETDATA;
		}
		// default value is null
		if (iSQLLength == -1)
			szDefValue [0] = '\0';


		// get SQL_DATETIME or SQL_INTERVAL
		SQLError = SQLGetData (hstmt, 14, SQL_C_SSHORT,
							   &iColDateInt, 0, &iSQLLength);
		if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
		{
			BF_ErrorSet ("", "", "Column 14");
			BF_ErrorDisplay (BF_ERR_SQLGETDATA);
			return BF_ERR_SQLGETDATA;
		}

		if (iColDateInt == SQL_DATETIME || iColDateInt == SQL_INTERVAL)
		{
			// get SQL_DATETIME or SQL_INTERVAL
			SQLError = SQLGetData (hstmt, 15, SQL_C_SSHORT,
								   &iColDateSub, 0, &iSQLLength);
			if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
			{
				BF_ErrorSet ("", "", "Column 15");
				BF_ErrorDisplay (BF_ERR_SQLGETDATA);
				return BF_ERR_SQLGETDATA;
			}
		}


		// declare field
		iError = BF_FieldAdd (sBFName, szColName, sColType, lColSize,
								 szColType, bColNull, szColRemark, szDefValue, iColDecimal);
		if (iError != 0)
			return iError;

		// reset error message because BF_AddField erase it
		BF_ErrorSet (sBFName, "BF_SQLRecordDescriptionLoad", "");

		// read next line
		SQLError = SQLFetchScroll (hstmt, SQL_FETCH_NEXT, 0);
		if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO &&
			SQLError != SQL_NO_DATA)
		{
			BF_ErrorDisplay (BF_ERR_SQLFETCHSCROLL);
			return BF_ERR_SQLFETCHSCROLL;
		}

	}

	// end sql query
	SQLError = SQLCloseCursor (hstmt);
	if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
	{
			BF_ErrorDisplay (BF_ERR_SQLCLOSECURSOR);
			return BF_ERR_SQLCLOSECURSOR;
	}

	// reset internal statement handle
	TmpTable.SQLStmt = NULL;

	// save current structure
	BFTable [sBFName] = TmpTable;

	return 0;

}



//-----------------------------------------------------------------
//!    BF_SQLQueryDescriptionLoad : load fields description       !
//-----------------------------------------------------------------
//! sBFName : internal query name                                 !
//!																  !
//! return : error or 0 if none                                   !
//-----------------------------------------------------------------
//! call by BF_OpenTable                                          !
//! 1024 fields max                                               !
//-----------------------------------------------------------------
int BF_SQLQueryDescriptionLoad (const string sBFName)
{
	int iError;
	BF_STRUCT_TABLE TmpTable;
	BF_STRUCT_FIELD TmpField;
	int iOrder;


	// for error management
	BF_ErrorSet (sBFName, "BF_SQLQueryDescriptionLoad", "");
	

	// read table structure
	TmpTable = BF_TableGet (sBFName);
	if (TmpTable.sDbName == "")
	{
		BF_ErrorDisplay (BF_ERR_NOTFOUND);
		return BF_ERR_NOTFOUND;
	}

	// verify type
	if (TmpTable.sDbType == BF_DB_BDB)
	{
		// DataBase Type is bad
		BF_ErrorDisplay (BF_ERR_BADTYPE);
		return BF_ERR_BADTYPE;
	}


	// get all fields
	iOrder = 1;
	while (1)
	{
		// get field info and stop at end
		TmpField = BF_SQLQueryFieldGet (sBFName, iOrder);
		if (TmpField.sDbFieldRemark == "" || iOrder > 1024)
			break;

		// declare field
		iError = BF_FieldAdd (sBFName, TmpField.sDbFieldRemark,
							  TmpField.sInterFieldType, TmpField.iDbFieldLength,
							  TmpField.sDbFieldType, TmpField.bDbFieldNull,
							  "", "", TmpField.iDbFieldDecimal);
		if (iError != 0)
			return iError;

		iOrder++;
	}

	return 0;

}



//-----------------------------------------------------------------
//!    BF_FieldAdd : add one field to the table description       !
//-----------------------------------------------------------------
//! sBFName : internal DataBase name                              !
//! sBFFieldName : internal Field name                            !
//! sBFInterType : internal type (ie : string, integer, ...)      !
//! iBFFieldLength : length  of field (>1) = size in bytes        !
//! sBFDbType : database type (ie : varchar, smallint, ...)       !
//! bBFDbNull : true = accept null, false = no null value         !
//! sBFDbRemark : database remark for the field (default = "")    !
//! sBFDbDefValue : database default value (default = "")         !
//! iBFDbDecimal : database num of char after decimal point       !
//!					(default = 0)								  !
//!																  !
//! return : error or 0 if none                                   !
//-----------------------------------------------------------------
int BF_FieldAdd (const string sBFName, const string sBFFieldName,
					const string sBFInterType, const int iBFFieldLength,
					const string sBFDbType, const bool bBFDbNull,
					const string sBFDbRemark, const string sBFDbDefValue,
					const int iBFDbDecimal)
					// sBFDbRemark = "", sBFDbDefValue = "", iBFDbDecimal = 0
{
	int iOrder;
	int iIndex;
	BFFieldItDef BFFieldIt;
	string sTable;
	BF_STRUCT_TABLE TmpTable;
	BF_STRUCT_FIELD TmpField;

	// for error management
	BF_ErrorSet (sBFName, "BF_FieldAdd", sBFFieldName);


	// read table structure
	TmpTable = BF_TableGet (sBFName);
	if (TmpTable.sDbName == "")
	{
		BF_ErrorDisplay (BF_ERR_NOTFOUND);
		return BF_ERR_NOTFOUND;
	}

	// read field structure
	TmpField = BF_FieldGet (sBFName, sBFFieldName);
	if (TmpField.sInterFieldType != "")
	{
		BF_ErrorDisplay (BF_ERR_FIELDEXIST);
		return BF_ERR_FIELDEXIST;
	}

	// empty field structure
	TmpField = BF_FieldNullGet ();


	// verify field name
	if (sBFFieldName == "")
	{
		BF_ErrorSet ("", "", "Field name is missing");
		BF_ErrorDisplay (BF_ERR_BADFIELDNAME);
		return BF_ERR_BADFIELDNAME;
	}

	// verify field type
	if (sBFDbType == "")
	{
		BF_ErrorSet ("", "", "DabaBase field type is missing");
		BF_ErrorDisplay (BF_ERR_BADFIELDTYPE);
		return BF_ERR_BADFIELDTYPE;
	}

	// verify field length
	if (sBFInterType == BF_FLD_STRING && iBFFieldLength < 1 ||
	    sBFInterType == BF_FLD_CURRENCY && iBFFieldLength < 1)
	{
		BF_ErrorDisplay (BF_ERR_BADFIELDLENGTH);
		return BF_ERR_BADFIELDLENGTH;
	}


	// store field length
	TmpField.iDbFieldDecimal = 0;
	if (sBFInterType == BF_FLD_STRING)
	{
		TmpField.iDbFieldLength = iBFFieldLength;
	}
	else if (sBFInterType == BF_FLD_BYTE)
	{
		TmpField.iDbFieldLength = 1;
	}
	else if (sBFInterType == BF_FLD_INTEGER)
	{
		TmpField.iDbFieldLength = 2;
	}
	else if (sBFInterType == BF_FLD_LONG)
	{
		TmpField.iDbFieldLength = 4;
	}
	else if (sBFInterType == BF_FLD_LONGLONG)
	{
		TmpField.iDbFieldLength = 8;
	}
	else if (sBFInterType == BF_FLD_FLOAT)
	{
		TmpField.iDbFieldLength = 4;
	}
	else if (sBFInterType == "memo")
	{
		TmpField.iDbFieldLength = 0;
	}
	else if (sBFInterType == BF_FLD_DATETIME)
	{
		TmpField.iDbFieldLength = 0;
	}
	else if (sBFInterType == BF_FLD_CURRENCY)
	{
		TmpField.iDbFieldLength = iBFFieldLength;
		TmpField.iDbFieldDecimal = iBFDbDecimal;
	}
	else if (sBFInterType == BF_FLD_BINARY)
	{
		TmpField.iDbFieldLength = 0;
	}
	else
	{
		BF_ErrorSet ("", "", BFError.sParam + ";" + sBFInterType);
		BF_ErrorDisplay (BF_ERR_BADFIELDTYPE);
		return BF_ERR_BADFIELDTYPE;
	}

	// compute field order
	iOrder = 1;

	for (BFFieldIt = BFField.begin(); BFFieldIt != BFField.end(); BFFieldIt++)
	{
		// extract table name
		sTable = BFFieldIt->first;
		iIndex = sTable.find ('.', 0);
		if (iIndex > 1)
			sTable = sTable.substr (0, iIndex);

		if (sTable == sBFName)
			iOrder++;
	}

	// store other variables
	TmpField.sInterFieldType = sBFInterType;
	TmpField.iDbFieldOrder = iOrder;
	TmpField.sDbFieldType = sBFDbType;
	TmpField.bDbFieldNull = bBFDbNull;
	TmpField.sDbFieldRemark = sBFDbRemark;
	TmpField.sDbFieldDefValue = sBFDbDefValue;

	// save current structure
	BFField [sBFName + "." + sBFFieldName] = TmpField;

	return 0;

}



//-----------------------------------------------------------------
//!    BF_SQLQueryOpen : send a SQL query to the database engine  !
//-----------------------------------------------------------------
//! sBFQueryName : internal Query name                            !
//! sBFSQLQuery : SQL Query request                               !
//! sBFDbName : OS DataBase name                                  !
//! sBFSrvName : server name                                      !
//! sBFDbType : DataBase type                                     !
//! sBFParam : param of query                                     !
//!																  !
//! return : "" or error message                                  !
//-----------------------------------------------------------------
string BF_SQLQueryOpen (const string sBFQueryName, const string sBFSQLQuery,
					 const string sBFDbName, const string sBFSrvName,
					 const string sBFDbType, const string sBFParam)
				  // sBFParam = ""
{
	BF_STRUCT_TABLE TmpTable;
	SQLRETURN SQLError;
	string sString;
	int iError;
	SQLCHAR SQLState [5];
	SQLCHAR SQLMessage [100];


	// for error management
	BF_ErrorSet (sBFQueryName, "BF_SQLQueryOpen", "");


	// internal table name is empty
	if (sBFQueryName == "")
	{
		BF_ErrorDisplay (BF_ERR_NAME);
		return BF_IntToString (BF_ERR_NAME);
	}

	// verify if file name contain "." character
	iError = sBFQueryName.find ('.', 0);
	if (iError >= 0)
	{
		BF_ErrorDisplay (BF_ERR_NOPOINT);
		return BF_IntToString (BF_ERR_NOPOINT);
	}

	// query is empty
	if (sBFSQLQuery == "")
	{
		BF_ErrorDisplay (BF_ERR_QUERYNULL);
		return BF_IntToString (BF_ERR_QUERYNULL);
	}

	// verify database type
	if (sBFDbType == BF_DB_BDB)
	{
		// DataBase Type is bad
		BF_ErrorDisplay (BF_ERR_BADTYPE);
		return BF_IntToString (BF_ERR_BADTYPE);
	}


	// get table structure
	TmpTable = BF_TableGet (sBFQueryName);
	if (TmpTable.sDbName != "")
	{
		BF_ErrorDisplay (BF_ERR_ALROPEN);
		return BF_IntToString (BF_ERR_ALROPEN);
	}


	// connect to database
	sString = sBFParam + ";" + BF_PRM_QUERY + "=true";
	iError = BF_TableOpen (sBFQueryName, "", sBFDbName, sBFSrvName, sBFDbType, sString);
	if (iError != 0)
		return BF_IntToString (iError);


	// get table structure
	TmpTable = BF_TableGet (sBFQueryName);

	// open statement handle, if not already open
	if (TmpTable.SQLStmt == NULL)
	{
		// statement handle
		SQLError = SQLAllocHandle (SQL_HANDLE_STMT, TmpTable.SQLDbc, &TmpTable.SQLStmt);
		if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
		{
			BF_ErrorDisplay (BF_ERR_SQLALLOCSTATEMENT);
			return BF_IntToString (BF_ERR_SQLALLOCSTATEMENT);
		}
	}

	// execute SQL command.
	SQLError = SQLExecDirect (TmpTable.SQLStmt, (SQLCHAR*) sBFSQLQuery.c_str(),
								sBFSQLQuery.length());
	if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
	{
		// get descrition of error
		SQLError = SQLGetDiagRec (SQL_HANDLE_STMT, TmpTable.SQLStmt, 1, SQLState, NULL,
										SQLMessage, sizeof(SQLMessage), NULL);

		sString = (char*) SQLState;
		sString += ":";
		sString += (char*) SQLMessage;

		return sString;
	}

	// save current structure
	BFTable [sBFQueryName] = TmpTable;

	// get fields description
	iError = BF_SQLQueryDescriptionLoad (sBFQueryName);
	if (iError)
		return BF_IntToString (iError);


	return "";

}



//-----------------------------------------------------------------
//!    BF_SQLQueryFieldGet : return field info for SQL Query      !
//-----------------------------------------------------------------
//! sBFTable : name of the table                                  !
//! iBFOrder : order number of the field in the record            !
//!                                                               !
//! return : structure BF_STRUCT_FIELD, empty if error            !
//-----------------------------------------------------------------
BF_STRUCT_FIELD BF_SQLQueryFieldGet (const string sBFTable, const int iBFOrder)
{
	BF_STRUCT_FIELD TmpField;
	BF_STRUCT_TABLE TmpTable;
	SQLRETURN SQLError;
	char szSQLName [BF_MAX_FIELDSIZE];
	SQLSMALLINT iSQLName;
	SQLSMALLINT iSQLType;
	SQLUINTEGER iSQLLength;
	SQLSMALLINT iSQLPrecision;
	SQLSMALLINT iSQLNullable;


	// no error management


	// read table structure
	TmpTable = BF_TableGet (sBFTable);
	if (TmpTable.sDbName == "")
		return BF_FieldNullGet ();


	SQLError = SQLDescribeCol (TmpTable.SQLStmt, iBFOrder, (SQLCHAR *) szSQLName,
							   sizeof (szSQLName), &iSQLName, &iSQLType,
							   &iSQLLength, &iSQLPrecision, &iSQLNullable);
	if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
		return BF_FieldNullGet ();


	// fill field structure
	TmpField = BF_FieldNullGet ();

	switch (iSQLType)
	{
		case SQL_CHAR:
		{
			TmpField.sInterFieldType = BF_FLD_STRING;
			TmpField.sDbFieldType = "SQL_CHAR";
			TmpField.iDbFieldLength = iSQLLength;

			break;
		}
		case SQL_NUMERIC:
		{
			TmpField.sInterFieldType = BF_FLD_CURRENCY;
			TmpField.sDbFieldType = "SQL_NUMERIC";
			TmpField.iDbFieldLength = iSQLLength;
			TmpField.iDbFieldDecimal = iSQLPrecision;

			break;
		}
		case SQL_DECIMAL:
		{
			TmpField.sInterFieldType = BF_FLD_CURRENCY;
			TmpField.sDbFieldType = "SQL_DECIMAL";
			TmpField.iDbFieldLength = iSQLLength;
			TmpField.iDbFieldDecimal = iSQLPrecision;

			break;
		}
		case SQL_INTEGER:
		{
			TmpField.sInterFieldType = BF_FLD_LONG;
			TmpField.sDbFieldType = "SQL_INTEGER";
			TmpField.iDbFieldLength = iSQLLength;

			break;
		}
		case SQL_SMALLINT:
		{
			TmpField.sInterFieldType = BF_FLD_INTEGER;
			TmpField.sDbFieldType = "SQL_SMALLINT";
			TmpField.iDbFieldLength = iSQLLength;

			break;
		}
		case SQL_FLOAT:
		{
			TmpField.sInterFieldType = BF_FLD_FLOAT;
			TmpField.sDbFieldType = "SQL_FLOAT";
			TmpField.iDbFieldLength = iSQLLength;

			break;
		}
		case SQL_REAL:
		{
			TmpField.sInterFieldType = BF_FLD_FLOAT;
			TmpField.sDbFieldType = "SQL_REAL";
			TmpField.iDbFieldLength = iSQLLength;

			break;
		}
		case SQL_DOUBLE:
		{
			TmpField.sInterFieldType = BF_FLD_FLOAT;
			TmpField.sDbFieldType = "SQL_DOUBLE";
			TmpField.iDbFieldLength = iSQLLength;

			break;
		}
		case SQL_VARCHAR:
		{
			TmpField.sInterFieldType = BF_FLD_STRING;
			TmpField.sDbFieldType = "SQL_VARCHAR";
			TmpField.iDbFieldLength = iSQLLength;

			break;
		}
		case SQL_TYPE_DATE:
		{
			TmpField.sInterFieldType = BF_FLD_DATETIME;
			TmpField.sDbFieldType = "SQL_TYPE_DATE";
			TmpField.iDbFieldLength = iSQLLength;

			break;
		}
		case SQL_TYPE_TIME:
		{
			TmpField.sInterFieldType = BF_FLD_DATETIME;
			TmpField.sDbFieldType = "SQL_TYPE_TIME";
			TmpField.iDbFieldLength = iSQLLength;

			break;
		}
		case SQL_TYPE_TIMESTAMP:
		{
			TmpField.sInterFieldType = BF_FLD_DATETIME;
			TmpField.sDbFieldType = "SQL_TYPE_TIMESTAMP";
			TmpField.iDbFieldLength = iSQLLength;

			break;
		}
		case -1:	// memo-Jet or text-SQLServer
		{
			TmpField.sInterFieldType = BF_FLD_STRING;
			TmpField.sDbFieldType = "MEMO";

			// size is limited to 1024 char
			if (iSQLLength < 1024)
				TmpField.iDbFieldLength = iSQLLength;
			else
				TmpField.iDbFieldLength = 1024;

			break;
		}
		case -2:	// binary-SQLServer
		{
			TmpField.sInterFieldType = BF_FLD_BINARY;
			TmpField.sDbFieldType = "BINARY";
			TmpField.iDbFieldLength = 0;

			break;
		}
		case -3:	// varbinary-SQLServer
		{
			TmpField.sInterFieldType = BF_FLD_BINARY;
			TmpField.sDbFieldType = "VARBINARY";
			TmpField.iDbFieldLength = 0;

			break;
		}
		case -4:	// image-SQLServer
		{
			TmpField.sInterFieldType = BF_FLD_BINARY;
			TmpField.sDbFieldType = "IMAGE";
			TmpField.iDbFieldLength = 0;

			break;
		}
		case -5:	// bigint-MySQL
		{
			TmpField.sInterFieldType = BF_FLD_LONGLONG;
			TmpField.sDbFieldType = "BIGINT";
			TmpField.iDbFieldLength = iSQLLength;

			break;
		}
		case -6:	// byte-Jet or tinyint-SQLServer/MySQL
		{
			TmpField.sInterFieldType = BF_FLD_BYTE;
			TmpField.sDbFieldType = "BYTE";
			TmpField.iDbFieldLength = iSQLLength;

			break;
		}
		case -7:	// bit-SQLServer
		{
			TmpField.sInterFieldType = BF_FLD_INTEGER;
			TmpField.sDbFieldType = "BIT";
			TmpField.iDbFieldLength = iSQLLength;

			break;
		}
		case -8:	// nchar-SQLServer
		{
			TmpField.sInterFieldType = BF_FLD_STRING;
			TmpField.sDbFieldType = "NCHAR";
			TmpField.iDbFieldLength = iSQLLength;

			break;
		}
		case -9:	// nvarchar-SQLServer
		{
			TmpField.sInterFieldType = BF_FLD_STRING;
			TmpField.sDbFieldType = "NVARCHAR";
			TmpField.iDbFieldLength = iSQLLength;

			break;
		}
		case -10:	// ntext-SQLServer
		{
			TmpField.sInterFieldType = BF_FLD_STRING;
			TmpField.sDbFieldType = "NTEXT";

			// size is limited to 1024 char
			if (iSQLLength < 1024)
				TmpField.iDbFieldLength = iSQLLength;
			else
				TmpField.iDbFieldLength = 1024;

			break;
		}
		case -11:	// uniqueid-SQLServer
		{
			TmpField.sInterFieldType = BF_FLD_INTEGER;
			TmpField.sDbFieldType = "UNIQUEID";
			TmpField.iDbFieldLength = iSQLLength;

			break;
		}
		default:
		{
			// type of field unknown
			return BF_FieldNullGet ();
		}
	}


	// field name is empty
	if (iSQLName != 0)
		TmpField.sDbFieldRemark = szSQLName;	// use remark for name
	else
		TmpField.sDbFieldRemark = "col" + BF_IntToString (iBFOrder);
	TmpField.iDbFieldOrder = iBFOrder;			// database order number
	if (iSQLNullable == 1)
		TmpField.bDbFieldNull = true;			// database null allowed
	else
		TmpField.bDbFieldNull = false;



	return TmpField;
}



//-----------------------------------------------------------------
//!    BF_SQLConnectStringGet : create connect string             !
//-----------------------------------------------------------------
//! sBFDbName : OS DataBase name                                  !
//! sBFSrvName : server name                                      !
//! sBFDbType : DataBase type (ie : SQL = SQL Server)             !
//! sBFParam : param of query                                     !
//!																  !
//! return : string or "" if error                                !
//-----------------------------------------------------------------
string BF_SQLConnectStringGet (const string sBFDbName, const string sBFSrvName,
							   const string sBFDbType, const string sBFParam)
{
	string sDriver;
	string sPort;
	string sString;


	// Microsoft Access database
	if (sBFDbType == BF_DB_ACCESS)
	{
		// prepare connection string
		sDriver = "DRIVER={Microsoft Access Driver (*.mdb)};DBQ=" + sBFDbName;

		// get trusted connection parameter
		sString = BF_StringParamGetIntl (BF_PRM_TRUST, sBFParam);
		if (!BF_StringIsTrue (sString))
		{
			// get user parameter
			sString = BF_StringParamGetIntl (BF_PRM_USER, sBFParam);
			if (sString != "")
			{
				sDriver += ";UID=";
				sDriver += sString;

				// get user parameter
				sString = BF_StringParamGetIntl (BF_PRM_PSW, sBFParam);
				sDriver += ";PWD=";
				sDriver += sString;
			}
		}
	}
	// Microsoft SQLServer database
	else if (sBFDbType == BF_DB_SQLSERVER)
	{
		// prepare connection string
		sDriver = "Driver={SQL Server};Server=" + sBFSrvName +
				  ";Database=" + sBFDbName;

		// get trusted connection parameter
		sString = BF_StringParamGetIntl (BF_PRM_TRUST, sBFParam);
		if (BF_StringIsTrue (sString))
			sDriver += ";Trusted_Connection=yes";
		else
		{
			// get user parameter
			sString = BF_StringParamGetIntl (BF_PRM_USER, sBFParam);
			if (sString != "")
			{
				sDriver += ";UID=";
				sDriver += sString;

				// get user parameter
				sString = BF_StringParamGetIntl (BF_PRM_PSW, sBFParam);
				sDriver += ";PWD=";
				sDriver += sString;
			}
		}
	}
	// MySQL Server database v3.51
	else if (sBFDbType == BF_DB_MYSQL)
	{
		// get port parameter
		sString = BF_StringParamGetIntl (BF_PRM_PORT, sBFParam);
		if (sString != "")
			sPort = sString;
		else
			sPort = "3306";

		// prepare connection string
		sDriver = "Driver={MySQL ODBC 3.51 Driver};Server=" + sBFSrvName +
				  ";Port=" + sPort + 
				  ";Database=" + sBFDbName;

		// get trusted connection parameter
		sString = BF_StringParamGetIntl (BF_PRM_TRUST, sBFParam);
		if (BF_StringIsTrue (sString))
			sDriver += ";Trusted_Connection=yes";
		else
		{
			// get user parameter
			sString = BF_StringParamGetIntl (BF_PRM_USER, sBFParam);
			if (sString != "")
			{
				sDriver += ";UID=";
				sDriver += sString;

				// get user parameter
				sString = BF_StringParamGetIntl (BF_PRM_PSW, sBFParam);
				sDriver += ";PWD=";
				sDriver += sString;
			}
		}

		sDriver += ";option=3;";
	}
	else
		sDriver = "";

	return sDriver;
}



//-----------------------------------------------------------------
//!    BF_SQLCommand : send a SQL command to the database engine  !
//-----------------------------------------------------------------
//! sBFSQLCommand : SQL command (ie:CREATE..., DELETE...)         !
//! sBFDbName : OS DataBase name                                  !
//! sBFSrvName : server name                                      !
//! sBFDbType : DataBase type (ie : SQL = SQL Server)             !
//! sBFParam : param of query                                     !
//!																  !
//! return : "" or error text                                     !
//-----------------------------------------------------------------
string BF_SQLCommand (const string sBFSQLCommand, const string sBFDbName,
					  const string sBFSrvName, const string sBFDbType, const string sBFParam)
{
	SQLHENV SQLEnvir;						// environnement handle
	SQLHDBC SQLDbc;							// connection handle
	SQLHSTMT SQLStmt;						// statement handle
	string sDriver;
	string sString;
	string sReturn;
	SQLCHAR SQLState [5];
	SQLCHAR SQLMessage [100];
	SQLRETURN SQLError;


	// no error management

	sReturn = "";

	// query is empty
	if (sBFSQLCommand == "")
	{
		BF_ErrorDisplay (BF_ERR_BADCOMMAND);
		return BF_IntToString (BF_ERR_BADCOMMAND);
	}

	// SQL database
	if (sBFDbType != BF_DB_BDB)
	{
		// environment handle
		SQLError = SQLAllocHandle (SQL_HANDLE_ENV, SQL_NULL_HANDLE, &SQLEnvir);
		if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
		{
			sReturn = BF_IntToString (BF_ERR_SQLALLOCENV);
			goto BF_SQLCommandSqlExit;
		}

		// environment attribute
		SQLError = SQLSetEnvAttr (SQLEnvir, SQL_ATTR_ODBC_VERSION,
									  (SQLPOINTER) SQL_OV_ODBC3, SQL_IS_INTEGER);
		if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
		{
			sReturn = BF_IntToString (BF_ERR_SQLSETENVATTR);
			goto BF_SQLCommandSqlExit;
		}

		// connection handle
		SQLError = SQLAllocHandle (SQL_HANDLE_DBC, SQLEnvir, &SQLDbc);
		if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
		{
			sReturn = BF_IntToString (BF_ERR_SQLSETENVATTR);
			goto BF_SQLCommandSqlExit;
		}

		// connection attribute
		SQLError = SQLSetConnectAttr (SQLDbc, SQL_ATTR_ACCESS_MODE,
										   (SQLPOINTER) SQL_MODE_READ_WRITE, 0);
		if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
		{
			sReturn = BF_IntToString (BF_ERR_SQLSETCONNATTR);
			goto BF_SQLCommandSqlExit;
		}


		// prepare connection string
		sDriver = BF_SQLConnectStringGet (sBFDbName, sBFSrvName, sBFDbType, sBFParam);

		// connection to the driver
		SQLError = SQLDriverConnect (SQLDbc, NULL,
								(SQLCHAR*) sDriver.c_str(), sDriver.length (),
								0, 0, 0, SQL_DRIVER_COMPLETE_REQUIRED);
		if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
		{
			sReturn = BF_IntToString (BF_ERR_SQLDRIVERCONNECT);
			goto BF_SQLCommandSqlExit;
		}

		// open statement handle
		SQLError = SQLAllocHandle (SQL_HANDLE_STMT, SQLDbc, &SQLStmt);
		if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
		{
			sReturn = BF_IntToString (BF_ERR_SQLALLOCSTATEMENT);
			goto BF_SQLCommandSqlExit;
		}

		// execute SQL command.
		SQLError = SQLExecDirect (SQLStmt, (SQLCHAR*) sBFSQLCommand.c_str(),
									sBFSQLCommand.length());
		if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
		{
			// get descrition of error
			SQLError = SQLGetDiagRec (SQL_HANDLE_STMT, SQLStmt, 1, SQLState, NULL,
										SQLMessage, sizeof(SQLMessage), NULL);

			sReturn = (char*) SQLState;
			sReturn += ":";
			sReturn += (char*) SQLMessage;

			goto BF_SQLCommandSqlExit;
		}


BF_SQLCommandSqlExit:
		// close handles
		SQLFreeHandle (SQL_HANDLE_STMT, SQLStmt);
		SQLDisconnect (SQLDbc);
		SQLFreeHandle (SQL_HANDLE_DBC, SQLDbc);
		SQLFreeHandle (SQL_HANDLE_ENV, SQLEnvir);
	}


	return sReturn;

}



//-----------------------------------------------------------------
//!    BF_TableClose : Close a database table                     !
//-----------------------------------------------------------------
//! sBFName : internal file name                                  !
//!																  !
//! return : error or 0 if none                                   !
//-----------------------------------------------------------------
int BF_TableClose (const string sBFName)
{
	int iFound;
	string sTableName;
	string sMapTable;
	BF_STRUCT_TABLE TmpTable;
	BF_STRUCT_FIELD TmpField;
	BFFieldItDef BFFieldIt;


	// for error management
	BF_ErrorSet (sBFName, "BF_TableClose", "");

	
	// read table structure
	TmpTable = BF_TableGet (sBFName);
	if (TmpTable.sDbName == "")
	{
		BF_ErrorDisplay (BF_ERR_NOTFOUND);
		return BF_ERR_NOTFOUND;
	}

	
	sTableName = sBFName + ".";

	// delete field for selected table, one by one
	iFound = 1;
	while (iFound == 1)
	{
		iFound = 0;
		for (BFFieldIt = BFField.begin(); BFFieldIt != BFField.end(); BFFieldIt++)
		{
			sMapTable = BFFieldIt->first;
			if (sMapTable.substr (0, sTableName.length()) == sTableName)
			{
				BFField.erase (sMapTable);
				iFound = 1;
				break;
			}
		}
	}


	// close table
	if (TmpTable.sDbType == BF_DB_BDB)
	{
		// close stream
		BFStream [TmpTable.iStream].BFStream.close ();

		// reset error flag
		BFStream [TmpTable.iStream].BFStream.clear ();

		// reset fields
		BFStream [TmpTable.iStream].BFMode = "";
		BFStream [TmpTable.iStream].BFInUse = false;
	}
	else
	{
		SQLFreeHandle (SQL_HANDLE_STMT, TmpTable.SQLStmt);
		SQLFreeHandle (SQL_HANDLE_STMT, TmpTable.SQLRecordStmt);
		SQLDisconnect (TmpTable.SQLDbc);
		SQLFreeHandle (SQL_HANDLE_DBC, TmpTable.SQLDbc);
		SQLFreeHandle (SQL_HANDLE_ENV, TmpTable.SQLEnvir);
	}


	// delete table structure
	BFTable.erase (sBFName);

	return 0;

}



//-----------------------------------------------------------------
//!    BF_TableGet : return information of a table                !
//-----------------------------------------------------------------
//! sBFTable : name of the table                                  !
//!                                                               !
//! return : table structure or a NULL structure if fail          !
//-----------------------------------------------------------------
BF_STRUCT_TABLE BF_TableGet (const string sBFTable)
{
	BF_STRUCT_TABLE TmpTable;
	BFTableItDef BFTableIt;


	// no error management

	// search table
	BFTableIt = BFTable.find (sBFTable);
	if (BFTableIt != BFTable.end())
		TmpTable = BFTable [sBFTable];
	else
		TmpTable = BF_TableNullGet ();

	return TmpTable;

}



//-----------------------------------------------------------------
//!    BF_FieldGet : return information of a field                !
//-----------------------------------------------------------------
//! sBFTable : name of the table                                  !
//! sBFField : name of the field                                  !
//!                                                               !
//! return : field structure or a NULL structure if fail          !
//-----------------------------------------------------------------
BF_STRUCT_FIELD BF_FieldGet (const string sBFTable, const string sBFField)
{
	BF_STRUCT_FIELD TmpField;
	BFFieldItDef BFFieldIt;


	// no error management

	// search field
	BFFieldIt = BFField.find (sBFTable + '.' + sBFField);
	if (BFFieldIt != BFField.end())
		TmpField = BFField [sBFTable + '.' + sBFField];
	else
		TmpField = BF_FieldNullGet ();

	return TmpField;

}



//-----------------------------------------------------------------
//!    BF_FieldNameGet : return field name by order               !
//-----------------------------------------------------------------
//! sBFTable : name of the table                                  !
//! iBFOrder : order number of the field in the record            !
//!                                                               !
//! return : field name or a "" if not found                      !
//-----------------------------------------------------------------
string BF_FieldNameGet (const string sBFTable, const int iBFOrder)
{
	string sTable;
	string sField;
	int iIndex;
	BF_STRUCT_FIELD TmpField;
	BFFieldItDef BFFieldIt;


	// no error management

	// search menu by order
	for (BFFieldIt = BFField.begin(); BFFieldIt != BFField.end(); BFFieldIt++)
	{
		// extract table name
		sTable = BFFieldIt->first;
		iIndex = sTable.find ('.', 0);
		if (iIndex < 0)
			break;

		sField = sTable.substr (iIndex + 1, sTable.length());
		sTable = sTable.substr (0, iIndex);

		TmpField = BFFieldIt->second;
		if (sTable == sBFTable && TmpField.iDbFieldOrder == iBFOrder)
			return sField;
	}

	return "";

}



//-----------------------------------------------------------------
//!    BF_TableNullGet : return null table structure              !
//-----------------------------------------------------------------
//! return : null BF_STRUCT_TABLE structure                       !
//-----------------------------------------------------------------
BF_STRUCT_TABLE BF_TableNullGet (void)
{
	BF_STRUCT_TABLE TmpTable;

	// no error management

	TmpTable.sDbName = "";
	TmpTable.sDbTableName = "";
	TmpTable.sDbType = "";
	TmpTable.sDbServer = "";
	TmpTable.SQLEnvir = NULL;
	TmpTable.SQLDbc = NULL;
	TmpTable.SQLStmt = NULL;
	TmpTable.SQLRecordStmt = NULL;
	TmpTable.iStream = -1;

	return TmpTable;
}



//-----------------------------------------------------------------
//!    BF_FieldNullGet : return null field structure              !
//-----------------------------------------------------------------
//! return : null BF_STRUCT_FIELD structure                       !
//-----------------------------------------------------------------
BF_STRUCT_FIELD BF_FieldNullGet (void)
{
	BF_STRUCT_FIELD TmpField;

	// no error management

	TmpField.sInterFieldType = "";
	TmpField.sDbFieldType = "";
	TmpField.iDbFieldOrder = 0;
	TmpField.iDbFieldLength = 0;
	TmpField.bDbFieldNull = false;
	TmpField.sDbFieldRemark = "";
	TmpField.sDbFieldDefValue = "";
	TmpField.iDbFieldDecimal = 0;
	TmpField.bIsNull = true;
	TmpField.sValue = "";
	TmpField.lValue = 0;

	return TmpField;
}



//-----------------------------------------------------------------
//!    BF_RecordReadNext : Read next record                       !
//-----------------------------------------------------------------
//! sBFName : internal file name                                  !
//!																  !
//! return : error or 0 if none                                   !
//!          -1 at the end of the file                            !
//-----------------------------------------------------------------
int BF_RecordReadNext (const string sBFName)
{
	int iFieldCounter;
	string sLine;
	string sString;
	SQLRETURN SQLError;
	BF_STRUCT_TABLE TmpTable;
	BF_STRUCT_FIELD TmpField;


	// for error management
	BF_ErrorSet (sBFName, "BF_RecordReadNext", "");


	// read table structure
	TmpTable = BF_TableGet (sBFName);
	if (TmpTable.sDbName == "")
	{
		BF_ErrorDisplay (BF_ERR_NOTFOUND);
		return BF_ERR_NOTFOUND;
	}

	
	// read one record
	if (TmpTable.sDbType == BF_DB_BDB)
	{
		// search for begin of record
		do
			getline (BFStream [TmpTable.iStream].BFStream, sLine);
		while (sLine != BF_BDB_BEGRECORD && !BFStream [TmpTable.iStream].BFStream.eof());

		// stop if end of file
		if (BFStream [TmpTable.iStream].BFStream.eof() == true)
			return -1;


		// reset field counter
		iFieldCounter = 1;
		
		// read all fields of the record
		do
		{
			// read next line
			getline (BFStream [TmpTable.iStream].BFStream, sLine);
			
			//begin-field and end-field found
			if (sLine [0] == BF_BDB_BEGFIELD && sLine [sLine.length () - 1] == BF_BDB_ENDFIELD)
			{
				// retrieve field number
				sString = BF_FieldNameGet (sBFName, iFieldCounter);

				// read field structure
				TmpField = BF_FieldGet (sBFName, sString);

				// store value 
				TmpField.sDbFieldRemark = sLine.substr (1, sLine.length() - 2);

				// save current structure
				BFField [sBFName + "." + sString] = TmpField;

				iFieldCounter++;
			}
		}
		while (sLine != BF_BDB_ENDRECORD && BFStream [TmpTable.iStream].BFStream.eof() == false);

		// end of file
		if (BFStream [TmpTable.iStream].BFStream.eof() == true)
			return -1;

	}
	else
	{
		// read one record
		SQLError = SQLFetchScroll (TmpTable.SQLStmt, SQL_FETCH_NEXT, 0);
		if (SQLError == SQL_NO_DATA)
		{
			// end of file
			return -1;
		}
		if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
		{
			BF_ErrorDisplay (BF_ERR_SQLFETCHSCROLL);
			return BF_ERR_SQLFETCHSCROLL;
		}
	}

	// save current structure
	BFTable [sBFName] = TmpTable;

	return 0;
}



//-----------------------------------------------------------------
//!    BF_FieldValueGetString : return a string field from record !
//-----------------------------------------------------------------
//! sBFName : internal file name                                  !
//! sBFField : internal field name                                !
//!																  !
//! return : value of the field                                   !
//!          or "" if error                                       !
//-----------------------------------------------------------------
string BF_FieldValueGetString (const string sBFName, const string sBFField)
{
	string sReturn;
	SQLRETURN SQLError;
	char szSQLChar [BF_MAX_FIELDSIZE];
	SQLINTEGER iSQLLength;
	BF_STRUCT_TABLE TmpTable;
	BF_STRUCT_FIELD TmpField;


	// for error management
	BF_ErrorSet (sBFName, "BF_FieldValueGetString", sBFField);


	// read table structure
	TmpTable = BF_TableGet (sBFName);
	if (TmpTable.sDbName == "")
	{
		BF_ErrorDisplay (BF_ERR_NOTFOUND);
		return "";
	}

	// read field structure
	TmpField = BF_FieldGet (sBFName, sBFField);
	if (TmpField.sInterFieldType == "")
	{
		BF_ErrorDisplay (BF_ERR_FIELDNOTFOUND);
		return "";
	}

	
	sReturn = "";

	
	// retrieve string
	if (TmpTable.sDbType == BF_DB_BDB)
	{
		sReturn = TmpField.sDbFieldRemark;
	}
	else
	{
		// retrieve one SQLCHAR field
		if (TmpField.sInterFieldType == BF_FLD_STRING || TmpField.sInterFieldType == "memo")
		{
			// retrieve one SQLCHAR field
			SQLError = SQLGetData (TmpTable.SQLStmt, TmpField.iDbFieldOrder, SQL_C_CHAR,
								   szSQLChar, sizeof (szSQLChar), &iSQLLength);
			if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
			{
				BF_ErrorDisplay (BF_ERR_SQLGETDATA);
				return "";
			}

			// string is not null
			if (iSQLLength != -1)
				sReturn = szSQLChar;

		}
		else
		{
			BF_ErrorDisplay (BF_ERR_FIELDNOTSTRING);
			return "";
		}
	}


	return sReturn;

}



//-----------------------------------------------------------------
//!    BF_FieldValueGetString : return a string field from record !
//-----------------------------------------------------------------
//! sBFName : internal file name                                  !
//! iBFOrder : order of the field                                 !
//!																  !
//! return : value of the field                                   !
//!          or "" if error                                       !
//-----------------------------------------------------------------
string BF_FieldValueGetString (const string sBFName, const int iBFOrder)
{
	string sString;


	// no error management


	// get field name
	sString = BF_FieldNameGet (sBFName, iBFOrder);

	return BF_FieldValueGetString (sBFName, sString);
}



//-----------------------------------------------------------------
//!    BF_FieldValueGetInteger : return a int field from record   !
//!            works for (unsigned byte, integer and long)        !
//-----------------------------------------------------------------
//! sBFName : internal file name                                  !
//! sBFField : internal field name                                !
//!																  !
//! return : value of the field                                   !
//!          or 0 if error                                        !
//-----------------------------------------------------------------
long BF_FieldValueGetInteger (const string sBFName, const string sBFField)
{
	long iLong;
	SQLRETURN SQLError;
	SQLINTEGER iSQLLength;
	char szSQLChar [BF_MAX_FIELDSIZE];
	SQLSMALLINT iSQLSmallInt;
	SQLINTEGER iSQLInteger;
	BF_STRUCT_TABLE TmpTable;
	BF_STRUCT_FIELD TmpField;


	// for error management
	BF_ErrorSet (sBFName, "BF_FieldValueGetInteger", sBFField);


	// read table structure
	TmpTable = BF_TableGet (sBFName);
	if (TmpTable.sDbName == "")
	{
		BF_ErrorDisplay (BF_ERR_NOTFOUND);
		return 0;
	}


	// read field structure
	TmpField = BF_FieldGet (sBFName, sBFField);
	if (TmpField.sInterFieldType == "")
	{
		BF_ErrorDisplay (BF_ERR_FIELDNOTFOUND);
		return 0;
	}



	// retrieve integer
	if (TmpTable.sDbType == BF_DB_BDB)
	{
		iLong = BF_StringToInt (TmpField.sDbFieldRemark);
	}
	else
	{
		// retrieve 1 byte unsigned integer 
		if (TmpField.sInterFieldType == BF_FLD_BYTE)
		{
			SQLError = SQLGetData (TmpTable.SQLStmt, TmpField.iDbFieldOrder, SQL_C_UTINYINT,
							   szSQLChar, sizeof(szSQLChar), &iSQLLength);
			if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
			{
				BF_ErrorDisplay (BF_ERR_SQLGETDATA);
				return 0;
			}

			// convert from signed char to unsigned int
			if (iSQLLength != -1)
			{
				iLong = szSQLChar [0];
				if (iLong < 0)
					iLong = 256 + iLong;
			}
			else
				// value is null
				iLong = 0;


		}
		// retrieve 2 bytes signed integer 
		else if (TmpField.sInterFieldType == BF_FLD_INTEGER)
		{
			SQLError = SQLGetData (TmpTable.SQLStmt, TmpField.iDbFieldOrder, SQL_C_SSHORT,
							   &iSQLSmallInt, sizeof(iSQLSmallInt), &iSQLLength);
			if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
			{
				BF_ErrorDisplay (BF_ERR_SQLGETDATA);
				return 0;
			}

			if (iSQLLength != -1)
				iLong = iSQLSmallInt;
			else
				// value is null
				iLong = 0;

		}
		// retrieve 4 bytes signed long integer 
		else if (TmpField.sInterFieldType == BF_FLD_LONG)
		{
			SQLError = SQLGetData (TmpTable.SQLStmt, TmpField.iDbFieldOrder, SQL_C_SLONG,
							   &iSQLInteger, sizeof(iSQLInteger), &iSQLLength);
			if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
			{
				BF_ErrorDisplay (BF_ERR_SQLGETDATA);
				return 0;
			}

			if (iSQLLength != -1)
				iLong = iSQLInteger;
			else
				// value is null
				iLong = 0;

		}
		else
		{
			BF_ErrorDisplay (BF_ERR_FIELDNOTINT);
			return 0;
		}
	}

	// save current structure
	BFField [sBFName + "." + sBFField] = TmpField;

	return iLong;

}



//-----------------------------------------------------------------
//!    BF_FieldValueGetInteger : return a int field from record   !
//!            works for (unsigned byte, integer and long)        !
//-----------------------------------------------------------------
//! sBFName : internal file name                                  !
//! iBFOrder : order of the field                                 !
//!																  !
//! return : value of the field                                   !
//!          or 0 if error                                        !
//-----------------------------------------------------------------
long BF_FieldValueGetInteger (const string sBFName, const int iBFOrder)
{
	string sString;


	// no error management


	// get field name
	sString = BF_FieldNameGet (sBFName, iBFOrder);

	return BF_FieldValueGetInteger (sBFName, sString);
}



//-----------------------------------------------------------------
//!    BF_FieldValueGetLongLong : return a int64 field from record!
//-----------------------------------------------------------------
//! sBFName : internal file name                                  !
//! sBFField : internal field name                                !
//!																  !
//! return : value of the field                                   !
//!          or 0 if error                                        !
//-----------------------------------------------------------------
LONGLONG BF_FieldValueGetLongLong (const string sBFName, const string sBFField)
{
	LONGLONG iLongLong;
	SQLRETURN SQLError;
	SQLINTEGER iSQLLength;
	SQLBIGINT iSQLLongLong;
	BF_STRUCT_TABLE TmpTable;
	BF_STRUCT_FIELD TmpField;


	// for error management
	BF_ErrorSet (sBFName, "BF_FieldValueGetLongLong", sBFField);


	// read table structure
	TmpTable = BF_TableGet (sBFName);
	if (TmpTable.sDbName == "")
	{
		BF_ErrorDisplay (BF_ERR_NOTFOUND);
		return 0;
	}


	// read field structure
	TmpField = BF_FieldGet (sBFName, sBFField);
	if (TmpField.sInterFieldType == "")
	{
		BF_ErrorDisplay (BF_ERR_FIELDNOTFOUND);
		return 0;
	}



	// retrieve longlong
	if (TmpTable.sDbType == BF_DB_BDB)
	{
		iLongLong = BF_StringToLongLong (TmpField.sDbFieldRemark);
	}
	else
	{
		if (TmpField.sInterFieldType == BF_FLD_LONGLONG)
		{
			SQLError = SQLGetData (TmpTable.SQLStmt, TmpField.iDbFieldOrder, SQL_C_SBIGINT,
							   &iSQLLongLong, sizeof(iSQLLongLong), &iSQLLength);
			if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
			{
				BF_ErrorDisplay (BF_ERR_SQLGETDATA);
				return 0;
			}

			if (iSQLLength != -1)
				iLongLong = iSQLLongLong;
			else
				// value is null
				iLongLong = 0;
		}
		else
		{
			BF_ErrorDisplay (BF_ERR_FIELDNOTINT);
			return 0;
		}
	}

	// save current structure
	BFField [sBFName + "." + sBFField] = TmpField;

	return iLongLong;

}



//-----------------------------------------------------------------
//!    BF_FieldValueGetLongLong : return a int64 field from record!
//-----------------------------------------------------------------
//! sBFName : internal file name                                  !
//! iBFOrder : order of the field                                 !
//!																  !
//! return : value of the field                                   !
//!          or 0 if error                                        !
//-----------------------------------------------------------------
LONGLONG BF_FieldValueGetLongLong (const string sBFName, const int iBFOrder)
{
	string sString;


	// no error management


	// get field name
	sString = BF_FieldNameGet (sBFName, iBFOrder);

	return BF_FieldValueGetLongLong (sBFName, sString);
}



//-----------------------------------------------------------------
//!    BF_FieldValueGetFloat : return a float field from record   !
//!            works for (float and double)                       !
//-----------------------------------------------------------------
//! sBFName : internal file name                                  !
//! sBFField : internal field name                                !
//!																  !
//! return : value of the field                                   !
//!          or 0 if error                                        !
//-----------------------------------------------------------------
double BF_FieldValueGetFloat (const string sBFName, const string sBFField)
{
	SQLREAL fSQLReel;
	SQLDOUBLE dSQLDouble;
	SQLRETURN SQLError;
	SQLINTEGER iSQLLength;
	double dDouble;
	BF_STRUCT_TABLE TmpTable;
	BF_STRUCT_FIELD TmpField;


	// for error management
	BF_ErrorSet (sBFName, "BF_FielValueGetFloat", sBFField);


	// read table structure
	TmpTable = BF_TableGet (sBFName);
	if (TmpTable.sDbName == "")
	{
		BF_ErrorDisplay (BF_ERR_NOTFOUND);
		return 0;
	}

	// read field structure
	TmpField = BF_FieldGet (sBFName, sBFField);
	if (TmpField.sInterFieldType == "")
	{
		BF_ErrorDisplay (BF_ERR_FIELDNOTFOUND);
		return 0;
	}


	if (TmpTable.sDbType == BF_DB_BDB)
	{
		dDouble = atof (TmpField.sDbFieldRemark.c_str ());
	}
	else
	{
		// retrieve 4 bytes float 
		if (TmpField.sInterFieldType == BF_FLD_FLOAT)
		{
			SQLError = SQLGetData (TmpTable.SQLStmt, TmpField.iDbFieldOrder, SQL_C_FLOAT,
							   &fSQLReel, sizeof(fSQLReel), &iSQLLength);
			if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
			{
				BF_ErrorDisplay (BF_ERR_SQLGETDATA);
				return 0;
			}

			dDouble = fSQLReel;
		}
		// retrieve 8 bytes double 
		else if (TmpField.sInterFieldType == "double")
		{
			SQLError = SQLGetData (TmpTable.SQLStmt, TmpField.iDbFieldOrder, SQL_C_DOUBLE,
							   &dSQLDouble, sizeof(dSQLDouble), &iSQLLength);
			if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
			{
				BF_ErrorDisplay (BF_ERR_SQLGETDATA);
				return 0;
			}

			dDouble = dSQLDouble;

		}
		else
		{
			BF_ErrorDisplay (BF_ERR_FIELDNOTINT);
			return 0;
		}
	}


	// save current structure
	BFField [sBFName + "." + sBFField] = TmpField;

	return dDouble;

}



//-----------------------------------------------------------------
//!    BF_FieldValueGetFloat : return a float field from record   !
//!            works for (float and double)                       !
//-----------------------------------------------------------------
//! sBFName : internal file name                                  !
//! iBFOrder : order of the field                                 !
//!																  !
//! return : value of the field                                   !
//!          or 0 if error                                        !
//-----------------------------------------------------------------
double BF_FieldValueGetFloat (const string sBFName, const int iBFOrder)
{
	string sString;


	// no error management


	// get field name
	sString = BF_FieldNameGet (sBFName, iBFOrder);

	return BF_FieldValueGetFloat (sBFName, sString);
}



//-----------------------------------------------------------------
//!    BF_FieldValueGetDate : return a datetime field from record !
//-----------------------------------------------------------------
//! sBFName : internal file name                                  !
//! sBFField : internal field name                                !
//!																  !
//! return : value of the field                                   !
//!          or null date if error                                !
//-----------------------------------------------------------------
BP_Date BF_FieldValueGetDate (const string sBFName, const string sBFField)
{
	BP_Date TempDate;
	BP_Date NullDate;
	SQLRETURN SQLError;
	SQLINTEGER iSQLLength;
	SQL_TIMESTAMP_STRUCT stSQLTimeStamp;
	SQL_TIME_STRUCT stSQLTime;
	BF_STRUCT_TABLE TmpTable;
	BF_STRUCT_FIELD TmpField;


	// for error management
	BF_ErrorSet (sBFName, "BF_FieldvalueGetDate", sBFField);


	// initialise null date for error return
	NullDate.setnull();

	// read table structure
	TmpTable = BF_TableGet (sBFName);
	if (TmpTable.sDbName == "")
	{
		BF_ErrorDisplay (BF_ERR_NOTFOUND);
		return NullDate;
	}

	// read field structure
	TmpField = BF_FieldGet (sBFName, sBFField);
	if (TmpField.sInterFieldType == "")
	{
		BF_ErrorDisplay (BF_ERR_FIELDNOTFOUND);
		return NullDate;
	}


	// retrieve date 
	if (TmpTable.sDbType == BF_DB_BDB)
	{
		TempDate.set (TmpField.sDbFieldRemark);
		return TempDate;
	}
	else
	{
		if (TmpField.sInterFieldType == BF_FLD_DATETIME)
		{
			// date or date-time field
			if (TmpField.sDbFieldType != "SQL_TYPE_TIME")
			{
				SQLError = SQLGetData (TmpTable.SQLStmt, TmpField.iDbFieldOrder,
						SQL_C_TYPE_TIMESTAMP, &stSQLTimeStamp, sizeof(stSQLTimeStamp),
						&iSQLLength);
				if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
				{
					BF_ErrorDisplay (BF_ERR_SQLGETDATA);
					return NullDate;
				}

				// convert date
				if (iSQLLength != -1)
				{
					TempDate.setdatetime (stSQLTimeStamp.year, stSQLTimeStamp.month,
							stSQLTimeStamp.day, stSQLTimeStamp.hour, stSQLTimeStamp.minute,
							stSQLTimeStamp.second);
				}
				else
				{
					// date is empty
					TempDate = NullDate;
				}
			}
			else
			{
				SQLError = SQLGetData (TmpTable.SQLStmt, TmpField.iDbFieldOrder,
						SQL_C_TYPE_TIME, &stSQLTime, sizeof(stSQLTime), &iSQLLength);
				if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
				{
					BF_ErrorDisplay (BF_ERR_SQLGETDATA);
					return NullDate;
				}

				// convert time
				if (iSQLLength != -1)
				{
					TempDate.setdatetime (1970, 1, 1, stSQLTime.hour, stSQLTime.minute,
							stSQLTime.second);
				}
				else
				{
					// date is empty
					TempDate = NullDate;
				}
			}
		}
		else
		{
			BF_ErrorDisplay (BF_ERR_FIELDNOTDATE);
			return NullDate;
		}
	}


	// save current structure
	BFField [sBFName + "." + sBFField] = TmpField;

	return TempDate;

}



//-----------------------------------------------------------------
//!    BF_FieldValueGetDate : return a datetime field from record !
//-----------------------------------------------------------------
//! sBFName : internal file name                                  !
//! iBFOrder : order of the field                                 !
//!																  !
//! return : value of the field                                   !
//!          or null date if error                                !
//-----------------------------------------------------------------
BP_Date BF_FieldValueGetDate (const string sBFName, const int iBFOrder)
{
	string sString;


	// no error management


	// get field name
	sString = BF_FieldNameGet (sBFName, iBFOrder);

	return BF_FieldValueGetDate (sBFName, sString);
}



//-----------------------------------------------------------------
//!    BF_FieldValueGetCurrency : return a currency field         !
//!                                       from a record           !
//-----------------------------------------------------------------
//! sBFName : internal file name                                  !
//! sBFField : internal field name                                !
//!																  !
//! return : value of the field                                   !
//!          or null currency if error                            !
//-----------------------------------------------------------------
BP_Currency BF_FieldValueGetCurrency (const string sBFName, const string sBFField)
{
	LONGLONG iLongLong;
	int iTemp;
	int iIndex;
	LONGLONG iHex;
	BP_Currency TempCur;
	BP_Currency NullCur;
	SQLRETURN SQLError;
	SQLINTEGER iSQLLength;
	SQL_NUMERIC_STRUCT nSQLNumeric;
	char szSQLChar [BF_MAX_FIELDSIZE];
	char c255 [256];
	string sString;
	string sIntPart;
	string sDecPart;
	BP_STRUCT_FORMAT BPFormat;
	BF_STRUCT_TABLE TmpTable;
	BF_STRUCT_FIELD TmpField;
	SQLHDESC hdesc;		// = NULL;


	// for error management
	BF_ErrorSet (sBFName, "BF_FieldValueGetCurrency", sBFField);


	// initialise null currency for error return
	NullCur.reset();

	// read table structure
	TmpTable = BF_TableGet (sBFName);
	if (TmpTable.sDbName == "")
	{
		BF_ErrorDisplay (BF_ERR_NOTFOUND);
		return NullCur;
	}

	// read field structure
	TmpField = BF_FieldGet (sBFName, sBFField);
	if (TmpField.sInterFieldType == "")
	{
		BF_ErrorDisplay (BF_ERR_FIELDNOTFOUND);
		return NullCur;
	}



	// retrieve currency
	if (TmpTable.sDbType == BF_DB_BDB)
	{
		TempCur.set (TmpField.sDbFieldRemark);
		return TempCur;
	}
	// other databases
	else
	{
		if (TmpField.sInterFieldType == BF_FLD_CURRENCY)
		{
			// field is not a SQL_DECIMAL
			if (TmpField.sDbFieldType != "SQL_DECIMAL")
			{
				// define numeric format
				SQLError = SQLGetStmtAttr (TmpTable.SQLStmt, SQL_ATTR_APP_ROW_DESC,
											&hdesc, 0, NULL);
				SQLError = SQLSetDescField (hdesc,TmpField.iDbFieldOrder, SQL_DESC_TYPE,
											(VOID*) SQL_C_NUMERIC, 0);
				SQLError = SQLSetDescField (hdesc, TmpField.iDbFieldOrder, SQL_DESC_PRECISION,
											(VOID*) TmpField.iDbFieldLength, 0);
				SQLError = SQLSetDescField (hdesc, TmpField.iDbFieldOrder, SQL_DESC_SCALE,
											(VOID*) TmpField.iDbFieldDecimal, 0);


				SQLError = SQLGetData (TmpTable.SQLStmt, TmpField.iDbFieldOrder, SQL_ARD_TYPE,
								   &nSQLNumeric, sizeof (nSQLNumeric), &iSQLLength);
				if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
				{
					BF_ErrorDisplay (BF_ERR_SQLGETDATA);
					return NullCur;
				}

				// convert currency
				if (iSQLLength > 0)
				{
					// convert SQL_C_NUMERIC to long long
					iLongLong = 0;
					iHex = 1;
					for (iIndex = 0; iIndex < SQL_MAX_NUMERIC_LEN; iIndex++)
					{
						iTemp = nSQLNumeric.val [iIndex];

						iLongLong += iHex * (iTemp % 16);
						iHex = iHex * 16;
						iLongLong += iHex * (iTemp / 16);
						iHex = iHex * 16;
					}

					// get current format
					BPFormat = TempCur.getformat();

					// if decimal sign don't exists, set to "."
					if (BPFormat.sBP_DecimalSign == "")
					{
						BPFormat.sBP_DecimalSign = ".";
						TempCur.setformat (BPFormat.iBP_NumInt,
										   BPFormat.iBP_NumDec,
										   BPFormat.sBP_DecimalSign,
										   BPFormat.sBP_KiloSign,
										   BPFormat.sBP_CurrencySymbol,
										   BPFormat.sBP_LeadChar,
										   BPFormat.sBP_Null,
										   false);
					}


					// integer part = 0
					iTemp = 1;
					sString = "";
					for (iIndex = 1; iIndex < nSQLNumeric.scale; iIndex++)
					{
						iTemp *= 10;
						sString += "0";
					}

					if (iLongLong < iTemp)
					{
						// convert long long to string
						_i64toa (iLongLong, c255, 10);
						sString += c255;

						if (nSQLNumeric.sign == 0)
							sString = "-0" + BPFormat.sBP_DecimalSign + sString;
						else
							sString = "0" + BPFormat.sBP_DecimalSign + sString;

					}
					else
					{
						if (nSQLNumeric.sign == 0)
							iLongLong *= -1;

						// convert long long to string
						_i64toa (iLongLong, c255, 10);
						sString = c255;

						// insert decimal sign
						sString.insert (sString.length() - nSQLNumeric.scale, ".");
					}

					// save value
					TempCur.set (sString);

				}
				// field is null
				else
					TempCur.setnull();

			}
			// field is SQL_DECIMAL
			else
			{
				// SQL_DECIMAL is store into char
				SQLError = SQLGetData (TmpTable.SQLStmt, TmpField.iDbFieldOrder, SQL_C_CHAR,
								   szSQLChar, sizeof (szSQLChar), &iSQLLength);
				if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
				{
					BF_ErrorDisplay (BF_ERR_SQLGETDATA);
					return NullCur;
				}

				TempCur.setstdformat (TmpField.iDbFieldLength, TmpField.iDbFieldDecimal, "US");
				TempCur.set (szSQLChar);
			}
		}
		// error field is not a currency
		else
		{
			BF_ErrorDisplay (BF_ERR_FIELDNOTCUR);
			return NullCur;
		}
	}


	// save current structure
	BFField [sBFName + "." + sBFField] = TmpField;

	return TempCur;

}



//-----------------------------------------------------------------
//!    BF_FieldValueGetCurrency : return a currency field         !
//!                                       from a record           !
//-----------------------------------------------------------------
//! sBFName : internal file name                                  !
//! iBFOrder : order of the field                                 !
//!																  !
//! return : value of the field                                   !
//!          or null currency if error                            !
//-----------------------------------------------------------------
BP_Currency BF_FieldValueGetCurrency (const string sBFName, const int iBFOrder)
{
	string sString;


	// no error management


	// get field name
	sString = BF_FieldNameGet (sBFName, iBFOrder);

	return BF_FieldValueGetCurrency (sBFName, sString);
}



//-----------------------------------------------------------------
//!    BF_ResetBipFILEIntl : Reset BipFILE structure              !
//-----------------------------------------------------------------
//! return : none                                                 !
//-----------------------------------------------------------------
//! INTERNAL FUNCTION                                             !
//-----------------------------------------------------------------
void BF_ResetBipFILEIntl (void)
{
	int iStream;


	// reset table structure
	BFTable.clear ();

	// reset field structure
	BFField.clear ();

	// reset stream structure
	for (iStream = 0; iStream < BF_MAX_STREAM; iStream++)
	{
		BFStream [iStream].BFInUse = false;
		BFStream [iStream].BFMode = "";
		BFStream [iStream].BFStream.close ();
		BFStream [iStream].BFStream.clear ();
	}

}



//-----------------------------------------------------------------
//!    BF_BDBRecordDescriptionLoad : load record description      !
//!          from BDB file                                        !
//-----------------------------------------------------------------
//! sBFName : internal file name                                  !
//!																  !
//! return : error or 0 if none                                   !
//-----------------------------------------------------------------
int BF_BDBRecordDescriptionLoad (const string sBFName)
{
	int iError;
	int iField;
	int iBegin;
	int iLength;
	string sLine;
	string sType;
	string sDbType;
	BF_STRUCT_TABLE TmpTable;


	// for error management
	BF_ErrorSet (sBFName, "BF_BDBRecordDescriptionLoad", "");
	
	// read table structure
	TmpTable = BF_TableGet (sBFName);
	if (TmpTable.sDbName == "")
	{
		BF_ErrorDisplay (BF_ERR_NOTFOUND);
		return BF_ERR_NOTFOUND;
	}

	// verify type
	if (TmpTable.sDbType != BF_DB_BDB)
	{
		// DataBase Type is bad
		BF_ErrorDisplay (BF_ERR_BADTYPE);
		return BF_ERR_BADTYPE;
	}

	// goto the beginning of the file
	BFStream [TmpTable.iStream].BFStream.seekg (0, std::ios::beg);
   

	// search for beginning of field description
	do
	{
		// read one line
		BFStream [TmpTable.iStream].BFStream >> sLine;

		// error, if found end description
		if (sLine == BF_BDB_ENDDESC)
		{
			BF_ErrorDisplay (BF_ERR_DESCKO);
			return BF_ERR_DESCKO;
		}
	}
	while (sLine != BF_BDB_BEGDESC && !ios::goodbit);

	// if eof or read error
	if (ios::goodbit)
	{
		BF_ErrorDisplay (BF_ERR_DESCNOTFOUND);
		return BF_ERR_DESCNOTFOUND;
	}


	// reset field index
	iField = 0;


	// load field description
	do
	{
		// read one line
		BFStream [TmpTable.iStream].BFStream >> sLine;

		// look only lines beginning with '<'
		if (sLine [0] == BF_BDB_BEGFIELD && sLine != BF_BDB_ENDDESC)
		{
			// last char must be '>'
			if (sLine [sLine.length() - 1] == BF_BDB_ENDFIELD)
			{
				// search for BDB description separator
				iBegin = sLine.find (BF_BDB_DESCSEPARATOR);

				// separator not found
				if (iBegin == 1)
				{
					BF_ErrorDisplay (BF_ERR_DESCNONSEPAR);
					return BF_ERR_DESCNONSEPAR;
				}


				sType = sLine.substr (iBegin + 1, sLine.length() - iBegin - 2);
//				sType = BF_StringToUpper (sType);
				if (sType == BF_FLD_STRING)
				{
					iLength = 255;
					sDbType = "varchar";
				}
				else if (sType == BF_FLD_INTEGER)
				{
					iLength = 2;
					sDbType = "smallint";
				}


				// store field into table
				iError = BF_FieldAdd (sBFName, sLine.substr (1, iBegin - 1), sType,
										 iLength, sDbType, true);
				if (iError != 0)
					return iError;

			}
			else
			{
				//malform descriptor
				BF_ErrorDisplay (BF_ERR_DESCNOENDFIELD);
				return BF_ERR_DESCNOENDFIELD;
			}
		}
	}
	while (sLine != BF_BDB_ENDDESC && !ios::goodbit);

	// if eof or read error
	if (ios::goodbit)
	{
		BF_ErrorDisplay (BF_ERR_EDESCNOTFOUND);
		return BF_ERR_EDESCNOTFOUND;
	}



	// return to the beginning of the file
	BFStream [TmpTable.iStream].BFStream.seekg (0, std::ios::beg);

	// save current structure
//	BFTable [sBFName] = TmpTable;

	return 0;

}



//-----------------------------------------------------------------
//!    BF_BDBRecordDescriptionSave : save record description      !
//!          into BDB file                                        !
//-----------------------------------------------------------------
//! sBFName : internal file name                                  !
//! sBFRecord : internal file name containing record description  !
//!																  !
//! return : error or 0 if none                                   !
//-----------------------------------------------------------------
int BF_BDBRecordDescriptionSave (const string sBFName, const string sBFRecord)
{
	int iError;
	int iFieldCounter;
	string sField;
	BF_STRUCT_FIELD TmpField;
	BF_STRUCT_TABLE TmpTable;
	BF_STRUCT_TABLE TmpRecord;


	// for error management
	BF_ErrorSet (sBFName, "BF_BDBRecordDescriptionSave", "");
	
	// read table structure
	TmpTable = BF_TableGet (sBFName);
	if (TmpTable.sDbName == "")
	{
		BF_ErrorDisplay (BF_ERR_NOTFOUND);
		return BF_ERR_NOTFOUND;
	}

	// read table structure
	TmpRecord = BF_TableGet (sBFRecord);
	if (TmpRecord.sDbName == "")
	{
		BF_ErrorDisplay (BF_ERR_NOTFOUND);
		return BF_ERR_NOTFOUND;
	}

	// verify type
	if (TmpTable.sDbType != BF_DB_BDB)
	{
		// DataBase Type is bad
		BF_ErrorDisplay (BF_ERR_BADTYPE);
		return BF_ERR_BADTYPE;
	}

	// verify type
	if (TmpRecord.sDbType != BF_DB_BDB)
	{
		// DataBase Type is bad
		BF_ErrorDisplay (BF_ERR_BADTYPE);
		return BF_ERR_BADTYPE;
	}

	// verify open mode
	if (BFStream [TmpTable.iStream].BFMode != "OUTPUT")
	{
		// mode must be OUTPUT
		BF_ErrorDisplay (BF_ERR_BADMODE);
		return BF_ERR_BADMODE;
	}

	// goto the beginning of the file
//	BFStream [TmpTable.iStream].BFStream.seekg (0, std::ios::beg);
   

	// copy header
	iFieldCounter = 1;

	BFStream [TmpTable.iStream].BFStream << BF_BDB_BEGDESC << endl;

	while (true)
	{
		// retreive field in order
		sField = BF_FieldNameGet (sBFRecord, iFieldCounter);

		TmpField = BFField [sBFRecord + "." + sField];

		// no more field
		if (TmpField.sDbFieldType == "")
			break;

		// save on field
		BFStream [TmpTable.iStream].BFStream << BF_BDB_BEGFIELD;
		BFStream [TmpTable.iStream].BFStream << sField;
		BFStream [TmpTable.iStream].BFStream << ":";
		BFStream [TmpTable.iStream].BFStream << TmpField.sInterFieldType;
		BFStream [TmpTable.iStream].BFStream << BF_BDB_ENDFIELD << endl;

		// store field into table
		if (sBFName != sBFRecord)
		{
		iError = BF_FieldAdd (sBFName, sField, TmpField.sInterFieldType,
								 TmpField.iDbFieldLength, TmpField.sDbFieldType, true);
		}

		iFieldCounter++;
	}

	BFStream [TmpTable.iStream].BFStream << BF_BDB_ENDDESC << endl << endl;


	return 0;

}



//-----------------------------------------------------------------
//!    BF_FieldSetString : set a string field into record         !
//-----------------------------------------------------------------
//! sBFName : internal file name                                  !
//! sBFField : internal field name                                !
//! sBFValue : value for the field                                !
//! bBFNull : true : field is null (default = false)              !
//!																  !
//! return : error or 0 if none                                   !
//-----------------------------------------------------------------
int BF_FieldSetString (const string sBFName, const string sBFField,
					   const string sBFValue, const bool bBFNull)
					   //const string sBFNull = false
{
	BF_STRUCT_TABLE TmpTable;
	BF_STRUCT_FIELD TmpField;


	// for error management
	BF_ErrorSet (sBFName, "BF_FieldSetString", sBFField);


	// read table structure
	TmpTable = BF_TableGet (sBFName);
	if (TmpTable.sDbName == "")
	{
		BF_ErrorDisplay (BF_ERR_NOTFOUND);
		return BF_ERR_NOTFOUND;
	}

	// read field structure
	TmpField = BF_FieldGet (sBFName, sBFField);
	if (TmpField.sInterFieldType == "")
	{
		BF_ErrorDisplay (BF_ERR_FIELDNOTFOUND);
		return BF_ERR_FIELDNOTFOUND;
	}

	// retrieve string
	if (TmpTable.sDbType == BF_DB_BDB)
	{
		// store in remark because it is free
		if (bBFNull)
			TmpField.sDbFieldRemark = "";
		else
			TmpField.sDbFieldRemark = sBFValue;
	}
	else
	{
		if (bBFNull)
		{
			TmpField.sValue = sBFValue;
			TmpField.bIsNull = true;
		}
		else
		{
			TmpField.sValue = sBFValue;
			TmpField.bIsNull = false;
		}
	}

	// save current structure
	BFField [sBFName + "." + sBFField] = TmpField;

	return 0;

}



//-----------------------------------------------------------------
//!    BF_FieldSetInteger : save a integer field into record      !
//-----------------------------------------------------------------
//! sBFName : internal file name                                  !
//! sBFField : internal field name                                !
//! lBFValue : value for the field                                !
//! bBFNull : true : field is null (default = false)              !
//!																  !
//! return : error or 0 if none                                   !
//-----------------------------------------------------------------
int BF_FieldSetInteger (const string sBFName, const string sBFField,
						const long lBFValue, const bool bBFNull)
						// const bool bBFNull = false
{
	BF_STRUCT_TABLE TmpTable;
	BF_STRUCT_FIELD TmpField;


	// for error management
	BF_ErrorSet (sBFName, "BF_FieldSetInteger", sBFField);


	// read table structure
	TmpTable = BF_TableGet (sBFName);
	if (TmpTable.sDbName == "")
	{
		BF_ErrorDisplay (BF_ERR_NOTFOUND);
		return BF_ERR_NOTFOUND;
	}

	// read field structure
	TmpField = BF_FieldGet (sBFName, sBFField);
	if (TmpField.sInterFieldType == "")
	{
		BF_ErrorDisplay (BF_ERR_FIELDNOTFOUND);
		return BF_ERR_FIELDNOTFOUND;
	}


	
	// retrieve string
	if (TmpTable.sDbType == BF_DB_BDB)
	{
		// store in remark because it is free
		if (bBFNull)
			TmpField.sDbFieldRemark = "";
		else
			TmpField.sDbFieldRemark = BF_IntToString (lBFValue);
	}
	else
	{
		// verify record state
		if (TmpTable.SQLRecordStmt == NULL)
		{
			BF_ErrorDisplay (BF_ERR_NOPREPARE);
			return BF_ERR_NOPREPARE;
		}

		if (TmpField.sInterFieldType == BF_FLD_INTEGER)
		{
			if (bBFNull)
			{
				TmpField.lValue = 0;
				TmpField.bIsNull = true;
			}
			else
			{
				TmpField.lValue = lBFValue;
				TmpField.bIsNull = false;
			}
		}
		else
		{
			BF_ErrorDisplay (BF_ERR_FIELDNOTINT);
			return BF_ERR_FIELDNOTINT;
		}
	}


	// save current structure
	BFField [sBFName + "." + sBFField] = TmpField;

	return 0;

}



//-----------------------------------------------------------------
//!    BF_FieldSetLongLong : save a int64 field into record       !
//-----------------------------------------------------------------
//! sBFName : internal file name                                  !
//! sBFField : internal field name                                !
//! llBFValue : value for the field                               !
//! bBFNull : true : field is null (default = false)              !
//!																  !
//! return : error or 0 if none                                   !
//-----------------------------------------------------------------
int BF_FieldSetLongLong (const string sBFName, const string sBFField,
						 const LONGLONG llBFValue, const bool bBFNull)
						 // const bool bBFNull = false
{
	BF_STRUCT_TABLE TmpTable;
	BF_STRUCT_FIELD TmpField;


	// for error management
	BF_ErrorSet (sBFName, "BF_FieldSetLongLong", sBFField);


	// read table structure
	TmpTable = BF_TableGet (sBFName);
	if (TmpTable.sDbName == "")
	{
		BF_ErrorDisplay (BF_ERR_NOTFOUND);
		return BF_ERR_NOTFOUND;
	}

	// read field structure
	TmpField = BF_FieldGet (sBFName, sBFField);
	if (TmpField.sInterFieldType == "")
	{
		BF_ErrorDisplay (BF_ERR_FIELDNOTFOUND);
		return BF_ERR_FIELDNOTFOUND;
	}


	
	// retrieve string
	if (TmpTable.sDbType == BF_DB_BDB)
	{
		// store in remark because it is free
		if (bBFNull)
			TmpField.sDbFieldRemark = "";
		else
			TmpField.sDbFieldRemark = BF_LongLongToString (llBFValue);
	}
	else
	{
		// verify record state
		if (TmpTable.SQLRecordStmt == NULL)
		{
			BF_ErrorDisplay (BF_ERR_NOPREPARE);
			return BF_ERR_NOPREPARE;
		}

		if (TmpField.sInterFieldType == BF_FLD_LONGLONG)
		{
			if (bBFNull)
			{
				TmpField.llValue = 0;
				TmpField.bIsNull = true;
			}
			else
			{
				TmpField.llValue = llBFValue;
				TmpField.bIsNull = false;
			}
		}
		else
		{
			BF_ErrorDisplay (BF_ERR_FIELDNOTINT);
			return BF_ERR_FIELDNOTINT;
		}
	}


	// save current structure
	BFField [sBFName + "." + sBFField] = TmpField;

	return 0;

}



//-----------------------------------------------------------------
//!    BF_FieldSetFloat : save a double field into record         !
//-----------------------------------------------------------------
//! sBFName : internal file name                                  !
//! sBFField : internal field name                                !
//! dBFValue : value for the field                                !
//! bBFNull : true : field is null (default = false)              !
//!																  !
//! return : error or 0 if none                                   !
//-----------------------------------------------------------------
int BF_FieldSetFloat (const string sBFName, const string sBFField,
					  const double dBFValue, const bool bBFNull)
					  // const bool bBFNull = false
{
	BF_STRUCT_TABLE TmpTable;
	BF_STRUCT_FIELD TmpField;


	// for error management
	BF_ErrorSet (sBFName, "BF_FieldSetFloat", sBFField);


	// read table structure
	TmpTable = BF_TableGet (sBFName);
	if (TmpTable.sDbName == "")
	{
		BF_ErrorDisplay (BF_ERR_NOTFOUND);
		return BF_ERR_NOTFOUND;
	}

	// read field structure
	TmpField = BF_FieldGet (sBFName, sBFField);
	if (TmpField.sInterFieldType == "")
	{
		BF_ErrorDisplay (BF_ERR_FIELDNOTFOUND);
		return BF_ERR_FIELDNOTFOUND;
	}


	
	// retrieve string
	if (TmpTable.sDbType == BF_DB_BDB)
	{
		// store in remark because it is free
		if (bBFNull)
			TmpField.sDbFieldRemark = "";
		else
			TmpField.sDbFieldRemark = BF_DoubleToString (dBFValue);
	}
	else
	{
		// verify record state
		if (TmpTable.SQLRecordStmt == NULL)
		{
			BF_ErrorDisplay (BF_ERR_NOPREPARE);
			return BF_ERR_NOPREPARE;
		}

		if (TmpField.sInterFieldType == BF_FLD_FLOAT)
		{
			if (bBFNull)
			{
				TmpField.dValue = 0;
				TmpField.bIsNull = true;
			}
			else
			{
				TmpField.dValue = dBFValue;
				TmpField.bIsNull = false;
			}
		}
		else
		{
			BF_ErrorDisplay (BF_ERR_FIELDNOTINT);
			return BF_ERR_FIELDNOTINT;
		}
	}


	// save current structure
	BFField [sBFName + "." + sBFField] = TmpField;

	return 0;

}



//-----------------------------------------------------------------
//!    BF_FieldSetDate : save a date field into record            !
//-----------------------------------------------------------------
//! sBFName : internal file name                                  !
//! sBFField : internal field name                                !
//! dtBFValue : value for the field                               !
//! bBFNull : true : field is null (default = false)              !
//!																  !
//! return : error or 0 if none                                   !
//-----------------------------------------------------------------
int BF_FieldSetDate (const string sBFName, const string sBFField,
					 BP_Date dtBFValue, const bool bBFNull)
					  // const bool bBFNull = false
{
	BF_STRUCT_TABLE TmpTable;
	BF_STRUCT_FIELD TmpField;


	// for error management
	BF_ErrorSet (sBFName, "BF_FieldSetDate", sBFField);


	// read table structure
	TmpTable = BF_TableGet (sBFName);
	if (TmpTable.sDbName == "")
	{
		BF_ErrorDisplay (BF_ERR_NOTFOUND);
		return BF_ERR_NOTFOUND;
	}

	// read field structure
	TmpField = BF_FieldGet (sBFName, sBFField);
	if (TmpField.sInterFieldType == "")
	{
		BF_ErrorDisplay (BF_ERR_FIELDNOTFOUND);
		return BF_ERR_FIELDNOTFOUND;
	}


	
	// retrieve string
	if (TmpTable.sDbType == BF_DB_BDB)
	{
		// store in remark because it is free
		if (bBFNull)
			TmpField.sDbFieldRemark = "";
		else
			TmpField.sDbFieldRemark = dtBFValue.str ();
	}
	else
	{
		// verify record state
		if (TmpTable.SQLRecordStmt == NULL)
		{
			BF_ErrorDisplay (BF_ERR_NOPREPARE);
			return BF_ERR_NOPREPARE;
		}

		if (TmpField.sInterFieldType == BF_FLD_DATETIME)
		{
			if (bBFNull)
			{
				TmpField.dtValue.setnull ();
				TmpField.bIsNull = true;
			}
			else
			{
				TmpField.dtValue = dtBFValue;
				TmpField.bIsNull = false;
			}
		}
		else
		{
			BF_ErrorDisplay (BF_ERR_FIELDNOTINT);
			return BF_ERR_FIELDNOTINT;
		}
	}


	// save current structure
	BFField [sBFName + "." + sBFField] = TmpField;

	return 0;

}



//-----------------------------------------------------------------
//!    BF_FieldSetCurrency : save a currency field into record    !
//-----------------------------------------------------------------
//! sBFName : internal file name                                  !
//! sBFField : internal field name                                !
//! curBFValue : value for the field                              !
//! bBFNull : true : field is null (default = false)              !
//!																  !
//! return : error or 0 if none                                   !
//-----------------------------------------------------------------
int BF_FieldSetCurrency (const string sBFName, const string sBFField,
					 BP_Currency curBFValue, const bool bBFNull)
					  // const bool bBFNull = false
{
	BF_STRUCT_TABLE TmpTable;
	BF_STRUCT_FIELD TmpField;


	// for error management
	BF_ErrorSet (sBFName, "BF_FieldSetDate", sBFField);


	// read table structure
	TmpTable = BF_TableGet (sBFName);
	if (TmpTable.sDbName == "")
	{
		BF_ErrorDisplay (BF_ERR_NOTFOUND);
		return BF_ERR_NOTFOUND;
	}

	// read field structure
	TmpField = BF_FieldGet (sBFName, sBFField);
	if (TmpField.sInterFieldType == "")
	{
		BF_ErrorDisplay (BF_ERR_FIELDNOTFOUND);
		return BF_ERR_FIELDNOTFOUND;
	}


	
	// retrieve string
	if (TmpTable.sDbType == BF_DB_BDB)
	{
		// store in remark because it is free
		if (bBFNull)
			TmpField.sDbFieldRemark = "";
		else
			TmpField.sDbFieldRemark = curBFValue.str ();
	}
	else
	{
		// verify record state
		if (TmpTable.SQLRecordStmt == NULL)
		{
			BF_ErrorDisplay (BF_ERR_NOPREPARE);
			return BF_ERR_NOPREPARE;
		}

		if (TmpField.sInterFieldType == BF_FLD_CURRENCY)
		{
			if (bBFNull)
			{
				TmpField.curValue.setnull ();
				TmpField.bIsNull = true;
			}
			else
			{
				TmpField.curValue = curBFValue;
				TmpField.bIsNull = false;
			}
		}
		else
		{
			BF_ErrorDisplay (BF_ERR_FIELDNOTINT);
			return BF_ERR_FIELDNOTINT;
		}
	}


	// save current structure
	BFField [sBFName + "." + sBFField] = TmpField;

	return 0;

}



//-----------------------------------------------------------------
//!    BF_RecordPrepare : prepare record before save              !
//-----------------------------------------------------------------
//! sBFName : internal file name                                  !
//!																  !
//! return : error or 0 if none                                   !
//-----------------------------------------------------------------
int BF_RecordPrepare (string sBFName)
{
	string sString;
	string sField;
	SQLRETURN SQLError;
	BF_STRUCT_TABLE TmpTable;
	BF_STRUCT_FIELD TmpField;
	BFFieldItDef BFFieldIt;


	// for error management
	BF_ErrorSet (sBFName, "BF_RecordPrepare", "");

	// get table structure
	TmpTable = BF_TableGet (sBFName);
	if (TmpTable.sDbName == "")
	{
		BF_ErrorDisplay (BF_ERR_NOTFOUND);
		return BF_ERR_NOTFOUND;
	}

	// statement handle
	SQLError = SQLAllocHandle (SQL_HANDLE_STMT, TmpTable.SQLDbc, &TmpTable.SQLRecordStmt);
	if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
	{
		BF_ErrorDisplay (BF_ERR_SQLALLOCSTATEMENT);
		return BF_ERR_SQLALLOCSTATEMENT;
	}



	// reset all field values
	for (BFFieldIt = BFField.begin(); BFFieldIt != BFField.end(); BFFieldIt++)
	{
		sField = BFFieldIt->first;
		TmpField = BFFieldIt->second;

		TmpField.bIsNull = true;
		TmpField.sValue = "";
		TmpField.lValue = 0;
		TmpField.llValue = 0;
		TmpField.dValue = 0;
		TmpField.dtValue.setnull ();
		TmpField.curValue.setnull ();

		// save field
		BFField [sField] = TmpField;
	}



	// save current structure
	BFTable [sBFName] = TmpTable;

	return 0;

}



//-----------------------------------------------------------------
//!    BF_RecordSave : save one record                            !
//-----------------------------------------------------------------
//! sBFName : internal file name                                  !
//!																  !
//! return : error or 0 if none                                   !
//-----------------------------------------------------------------
int BF_RecordSave (const string sBFName)
{
	int iFieldCounter;
	string sField;
	string sString;
	string sValues;
	BF_STRUCT_TABLE TmpTable;
	BF_STRUCT_FIELD TmpField;
	BFFieldItDef BFFieldIt;
	SQLRETURN SQLError;


	// for error management
	BF_ErrorSet (sBFName, "BF_RecordSave", "");


	// read table structure
	TmpTable = BF_TableGet (sBFName);
	if (TmpTable.sDbName == "")
	{
		BF_ErrorDisplay (BF_ERR_NOTFOUND);
		return BF_ERR_NOTFOUND;
	}


	// save one record
	if (TmpTable.sDbType == BF_DB_BDB)
	{
		// verify open mode
		if (BFStream [TmpTable.iStream].BFMode != "OUTPUT" &&
			BFStream [TmpTable.iStream].BFMode != "INOUT")
		{
			// mode must be OUTPUT
			BF_ErrorDisplay (BF_ERR_BADMODE);
			return BF_ERR_BADMODE;
		}

		// goto end of file
		BFStream [TmpTable.iStream].BFStream.seekg (0, std::ios::end);

		// save all fields
		iFieldCounter = 1;

		BFStream [TmpTable.iStream].BFStream << endl << BF_BDB_BEGRECORD << endl;

		while (true)
		{
			// retrieve field in order
			sField = BF_FieldNameGet (sBFName, iFieldCounter);

			TmpField = BF_FieldGet (sBFName, sField);

			// no more field
			if (TmpField.sDbFieldType == "")
				break;

			// save one field
			BFStream [TmpTable.iStream].BFStream << BF_BDB_BEGFIELD;
			BFStream [TmpTable.iStream].BFStream << TmpField.sDbFieldRemark;
			BFStream [TmpTable.iStream].BFStream << BF_BDB_ENDFIELD << endl;

			iFieldCounter++;
		}

		BFStream [TmpTable.iStream].BFStream << BF_BDB_ENDRECORD << endl << endl;

	}
	else
	{
		// prepare INSERT command : fields name
		sString = "INSERT INTO " + BF_BIP_SQLTABLE + " (";

		// prepare fields value
		sValues = ") VALUES (";

		// retrieve all fields, no order
		iFieldCounter = 1;

		for (BFFieldIt = BFField.begin(); BFFieldIt != BFField.end(); BFFieldIt++)
		{
			sField = BFFieldIt->first;
			sField = sField.substr (sBFName.length () + 1);
//			TmpField = BF_FieldGet (sBFName, sField);
			TmpField = BFFieldIt->second;

			if (iFieldCounter > 1)
			{
				sString += ",";
				sValues += ",";
			}

			// add field name
			sString += sField;

			// add field value
			TmpField = BF_FieldGet (sBFName, sField);

			// string
			if (TmpField.sInterFieldType == BF_FLD_STRING)
			{
				// null value
				if (TmpField.bIsNull ||
					(TmpField.sValue == "" && TmpField.bDbFieldNull))
					sValues += "NULL";
				else
					sValues += "'" + TmpField.sValue + "'";
			}
			// integer, long, byte
			else if (TmpField.sInterFieldType == BF_FLD_INTEGER
					 || TmpField.sInterFieldType == BF_FLD_LONG
					 || TmpField.sInterFieldType == BF_FLD_BYTE)
				sValues += BF_IntToString (TmpField.lValue);
			// longlong
			else if (TmpField.sInterFieldType == BF_FLD_LONGLONG)
				sValues += BF_LongLongToString (TmpField.llValue);
			// float, double
			else if (TmpField.sInterFieldType == BF_FLD_FLOAT)
				sValues += BF_DoubleToString (TmpField.dValue);
			else if (TmpField.sInterFieldType == BF_FLD_DATETIME)
				sValues += TmpField.dtValue.str ();
			else if (TmpField.sInterFieldType == BF_FLD_CURRENCY)
				sValues += TmpField.curValue.str ();

			iFieldCounter++;
		}

		// terminate INSERT command
		sString = sString + sValues += ")";


		// execute SQL command.
		SQLError = SQLExecDirect (TmpTable.SQLRecordStmt, (SQLCHAR*) sString.c_str(),
									sString.length());
		if (SQLError != SQL_SUCCESS && SQLError != SQL_SUCCESS_WITH_INFO)
		{
			BF_ErrorDisplay (BF_ERR_SQLEXECDIRECT);
			return BF_ERR_SQLEXECDIRECT;
		}

		// close statement
		SQLFreeHandle (SQL_HANDLE_STMT, TmpTable.SQLRecordStmt);
		TmpTable.SQLRecordStmt = NULL;

	}

	// save current structure
	BFTable [sBFName] = TmpTable;

	return 0;
}



//-----------------------------------------------------------------
//!    BF_StringParamGetIntl : return value of a parameter        !
//-----------------------------------------------------------------
//! sBFParam : name of the parameter                              !
//! sBFString : string containing all parameters                  !
//!                                                               !
//! return : value of the param or "" if error or not found       !
//-----------------------------------------------------------------
//! - search is not case sensitive                                !
//! - only the first occurence of the parameter is return         !
//! - format of sBF_String : "param1=value1;param2=value2; ..."   !
//!     - character between param and value -> "="                !
//!     - character between parameters -> ";"                     !
//-----------------------------------------------------------------
//!  INTERNAL FUNCTION                                            !
//-----------------------------------------------------------------
string BF_StringParamGetIntl (const string sBFParam, const string sBFString)
{
	int iBegin;
	int iEnd;
	string sParam;
	string sString;
	string sTemp;


	// no error management

	// end if one input string is empty
	if (sBFParam == "" || sBFString == "")
		return "";

	// tranfert into local and upper case string 
	sParam = BF_StringToUpper (sBFParam);
	sString = BF_StringToUpper (sBFString);

	// search for parameter
	iBegin = -1;
	do
		iBegin = sString.find (sParam + "=", iBegin + 1);
	while (iBegin > 0 && sString [iBegin - 1] != ';');

	if (iBegin < 0)
	{
		// param not found
		return "";
	}
	else
	{
		//select and return value
		iBegin += sParam.length ();
		iBegin += 1;

		iEnd = sString.find (";", iBegin);
		if (iEnd < 0)
			iEnd = sString.length () + iBegin;

		sTemp = sBFString.substr (iBegin, iEnd - iBegin);

		if (sTemp == "")
			return sTemp;

		// delete leading and ending spaces
		iBegin = sTemp.find_first_not_of (" ");
		iEnd = sTemp.find_last_not_of (" ");

		return sTemp.substr (iBegin, iEnd - iBegin + 1);
	}
}



//-----------------------------------------------------------------
//!    BF_StringIsTrue : tell if a string is consider as true     !
//-----------------------------------------------------------------
//! sBFInput : input string                                       !
//!                                                               !
//! return : true or false                                        !
//-----------------------------------------------------------------
//! string content is consider true, if it is equal to :          !
//!  (not case sensitive) true,t,yes,y                            !
//! in all other case (ie string is empty) content is consider    !
//!     false                                                     !
//-----------------------------------------------------------------
bool BF_StringIsTrue (const string sBFInput)
{
	string sString;

	// no error management

	sString = BF_StringToUpper (sBFInput);

	if (sString == "TRUE" || sString == "T" || sString == "YES" || sString == "Y")
		return true;
	else
		return false;

}



//-----------------------------------------------------------------
//!    BF_StringIsFalse : tell if a string is consider as false   !
//-----------------------------------------------------------------
//! sBFInput : input string                                       !
//!                                                               !
//! return : true or false                                        !
//-----------------------------------------------------------------
//! string content is consider false, if it is equal to :         !
//!  (not case sensitive) false,f,no,n                            !
//! in all other case (ie string is empty) content is consider    !
//!     false                                                     !
//-----------------------------------------------------------------
bool BF_StringIsFalse (const string sBFInput)
{
	string sString;

	// no error management

	sString = BF_StringToUpper (sBFInput);

	if (sString == "FALSE" || sString == "F" || sString == "NO" || sString == "N")
		return true;
	else
		return false;

}



//-----------------------------------------------------------------
//!    BF_ErrorSet : store error message                          !
//-----------------------------------------------------------------
//! sBFName : name of the BipFile object                          !
//! sBFFunction : name of the BipFile function                    !
//! sBFParam : name of the BipFile function                       !
//!																  !
//! return : none                                                 !
//-----------------------------------------------------------------
void BF_ErrorSet (const string sBFName, const string sBFFunction, const string sBFParam)
{

	if (sBFName != "")
		BFError.sName = sBFName;

	if (sBFFunction != "")
		BFError.sFunction = sBFFunction;

	BFError.sParam = sBFParam;

}



//-----------------------------------------------------------------
//!    BF_ErrorDisplaySet : modify the error display flag         !
//-----------------------------------------------------------------
//! bBFError : true  = display error message                      !
//!       default value is true                                   !
//! bBFWrite : true  = write error message in biperror.txt file   !
//!       default value is false                                  !
//-----------------------------------------------------------------
void BF_ErrorDisplaySet (const bool bBFError, const bool bBFWrite)
				// bBFError = true, bBFWrite = false
{

	if (bBFError == true)
		BF_DISPLAY_ERROR = true;
	else
		BF_DISPLAY_ERROR = false;

	if (bBFWrite == true)
		BF_WRITE_ERROR = true;
	else
		BF_WRITE_ERROR = false;
}



//-----------------------------------------------------------------
//!    BF_ErrorDisplay : display error when using BipFile         !
//-----------------------------------------------------------------
//! iBGError : error number                                       !
//!																  !
//! return : error message                                        !
//-----------------------------------------------------------------
//!	display error only if flag BF_DISPLAY_ERROR is true           !
//-----------------------------------------------------------------
string BF_ErrorDisplay (const int iBFError)
{
	int iReturn;
	string sMsg;
	string sDbName;
	string sTableName;
	string sDbType;
	SQLSMALLINT SQLHType;
	SQLHANDLE SQLHandle;
	SQLCHAR SQLState [5];
	SQLCHAR SQLMessage [100];
	SQLRETURN SQLError;
	string sString;
	ifstream ErrorStream;
	ofstream OutStream;
	string sErrorMsg;
	string sMsgType;
	string sMsgCat;
	string sMsgLang;
	int iMsgNo;
	string sMsgMsg;
	string sLanguage;
	char cChar20 [20];
	BF_STRUCT_TABLE TmpTable;


	// read table structure
	TmpTable = BF_TableGet (BFError.sName);

	// get table,file name
	if (TmpTable.sDbName == "")
	{
		sDbName = "none";
		sTableName = sDbName;
		sDbType = sDbName;
	}
	else
	{
		sDbName = TmpTable.sDbName;
		sTableName = TmpTable.sDbTableName;
		sDbType = TmpTable.sDbType;
	}
	
	// open file containing error message
	ErrorStream.open ("bipmsg.bdb", ios::in);
	if (ErrorStream.fail())
		sErrorMsg = "no message : file bipmsg.bdb not found";
	else
	{
		// select language
		sLanguage = BF_Language;

		// first search in default language. if not found, search in "US " language
		do
		{
			// select US language if default language not found
			if (sLanguage != "US" && ErrorStream.eof())
			{
				sLanguage = "US";

				// clear error state
				ErrorStream.clear();

				// return to the beginning of the file
				ErrorStream.seekg (0, std::ios::beg);
			}

			// search for error message
			do
			{
				// search for begin of record
				do
					getline (ErrorStream, sString);
				while (sString != BF_BDB_BEGRECORD && !ErrorStream.eof());

				// stop if end of file
				if (ErrorStream.eof() == true)
					break;


				// read next line (supposed to be type)
				getline (ErrorStream, sString);
				if (sString.length() > 2)
					sMsgType = sString.substr (1, sString.length() - 2);

				// read next line (supposed to be cat)
				getline (ErrorStream, sString);
				if (sString.length() > 2)
					sMsgCat = sString.substr (1, sString.length() - 2);

				// read next line (supposed to be lang)
				getline (ErrorStream, sString);
				if (sString.length() > 2)
					sMsgLang = sString.substr (1, sString.length() - 2);

				// read next line (supposed to be no)
				getline (ErrorStream, sString);
				if (sString.length() > 2)
					iMsgNo = BF_StringToInt(sString.substr (1, sString.length() - 2));

				// read next line (supposed to be msg)
				getline (ErrorStream, sString);
				if (sString.length() > 2)
					sMsgMsg = sString.substr (1, sString.length() - 2);

			}
			while (sMsgType != "ERROR" || sMsgCat != "BIPFILE" || sMsgLang != sLanguage
					|| iMsgNo != iBFError || ErrorStream.eof());

		}
		while ((sMsgType != "ERROR" || sMsgCat != "BIPFILE" || iMsgNo != iBFError)
			&& sLanguage != "US");


		if (ErrorStream.eof())
			sErrorMsg = "no message : message error not found";
		else
			sErrorMsg = sMsgMsg;

		// close message error file
		ErrorStream.close();
	
	}


	// contruct the message to display
	sMsg = "Function : " + BFError.sFunction + '\n' +
		   "Internal table name: " + BFError.sName + '\n' +
		   "DataBase Name : " + sDbName + '\n' +
		   "Table Name : " + sTableName + '\n' +
		   "DataBase Type : " + sDbType + '\n' +
		   "Error number : " + BF_IntToString(iBFError) + '\n' +
		   "Error msg : " + sErrorMsg + '\n';

	// display param if not empty
	if (BFError.sParam != "")
	   sMsg += "Param : " + BFError.sParam + '\n';

	// if odbc error : get odbc message
	if (iBFError >= 1000 && iBFError < 2000)
	{
		// select type of error
		SQLHType = SQL_HANDLE_ENV;
		SQLHandle = TmpTable.SQLEnvir;
		if (iBFError >= 1200 && iBFError < 1300)
		{
			SQLHType = SQL_HANDLE_DBC;
			SQLHandle = TmpTable.SQLDbc;
		}
		else if (iBFError >= 1300 && iBFError < 1400)
		{
			SQLHType = SQL_HANDLE_STMT;
			SQLHandle = TmpTable.SQLStmt;
		}

		// get odbc error
		SQLError = SQLGetDiagRec (SQLHType, SQLHandle, 1, SQLState, NULL,
					   SQLMessage, sizeof(SQLMessage), NULL);

		if (SQLError == SQL_SUCCESS || SQLError == SQL_SUCCESS_WITH_INFO)
		{
			sString = (char*) SQLState;
			sMsg += "ODBC state : " + sString + '\n';
			sString = (char*) SQLMessage;
			sMsg +=	"ODBC msg : " + sString + '\n';
		}
		else if (SQLError == SQL_INVALID_HANDLE)
		{
			sMsg += "ODBC msg : ";
			sMsg += "SQL_INVALID_HANDLE";
			sMsg += '\n';
		}
		else if (SQLError == SQL_ERROR)
		{
			sMsg += "ODBC msg : ";
			sMsg += "SQL_ERROR";
			sMsg += '\n';
		}
		else if (SQLError == SQL_NO_DATA)
		{
			sMsg += "ODBC msg : ";
			sMsg += "SQL_NO_DATA";
			sMsg += '\n';
		}
	}

	
	// display error if flag is set
	if (BF_DISPLAY_ERROR == true)
	{
		// contruct the message to display
		sMsg += '\n';
		sMsg += "Do you want to stop the program ?";

		// display a messagebox with error icon and 2 button (yes/no)
		iReturn = MessageBox(NULL, sMsg.data(), "BipFile Error", MB_YESNO | MB_ICONERROR);

		// stop if user clic on "Yes" button, else continue program
		if (iReturn == IDYES)
		{
			// end the program
			PostQuitMessage(0);
		}
	}
	else if (BF_WRITE_ERROR == true)
	{
		// open output file
		OutStream.open ("biperror.txt", ios::app);
		if (!OutStream.fail())
		{
			// write message to file
			_strdate (cChar20);
			OutStream << cChar20 << ' ';
			_strtime (cChar20);
			OutStream << cChar20 << endl;
			OutStream << sMsg << endl;
			OutStream.close();
		}
	}

	return sMsg;

}



//-----------------------------------------------------------------
//!    BF_IntToString : Convert an long integer into a string     !
//-----------------------------------------------------------------
//! iLong : int number to convert                                 !
//!                                                               !
//! return : number into a string                                 !
//-----------------------------------------------------------------
string BF_IntToString (const long iLong)
{
	ostringstream sStream;

	sStream << iLong;

	return sStream.str();
}



//-----------------------------------------------------------------
//!    BF_LongLongToString : Convert an integer64 into a string   !
//-----------------------------------------------------------------
//! iLong : LONGLONG number to convert                            !
//!                                                               !
//! return : number into a string                                 !
//-----------------------------------------------------------------
string BF_LongLongToString (const LONGLONG iLong)
{
	string sString;
	char cChar [50];

	_i64toa (iLong, cChar, 10);
	sString = cChar;

	return sString;
}



//-----------------------------------------------------------------
//!    BF_DoubleToString : Convert an double(float) into a string !
//-----------------------------------------------------------------
//! dDouble : double or float number to convert                   !
//!                                                               !
//! return : number into a string                                 !
//-----------------------------------------------------------------
string BF_DoubleToString (const double dDouble)
{
	ostringstream sStream;

	sStream << dDouble;

	return sStream.str();
}



//-----------------------------------------------------------------
//!    BF_StringToInt : Convert a string to long integer          !
//-----------------------------------------------------------------
//! sString : string containing integer value                     !
//! iBase : base of the number into the string (default = 10)     !
//!                                                               !
//! return : integer value                                        !
//-----------------------------------------------------------------
int BF_StringToInt (const string sString, const int iBase)
				// iBase = 10
{
	int iBGint;
	int iTemp;

	// return zero if string is empty
	if (sString == "")
		return 0;

	// use decimal if base is ko
	if (iBase != 10 && iBase != 16 && iBase != 8 && iBase != 2)
		iTemp = 10;
	else
		iTemp = iBase;

	istringstream sStream (sString);
	sStream >> setbase (iTemp) >> iBGint;

	return iBGint;

}



//-----------------------------------------------------------------
//!    BF_StringToLongLong : Convert a string to an integer64     !
//-----------------------------------------------------------------
//! sString : string containing integer value                     !
//!                                                               !
//! return : integer value                                        !
//-----------------------------------------------------------------
LONGLONG BF_StringToLongLong (const string sString)
{
	LONGLONG iBGlong;

	iBGlong = _atoi64 (sString.c_str());

	return iBGlong;
}



//########## MEN AT WORK ##################

//-----------------------------------------------------------------
//!    BF_StringToUpper : Convert a string to upper case          !
//-----------------------------------------------------------------
//! sString : String in lower case box                            !
//!                                                               !
//! return : String in upper case                                 !
//-----------------------------------------------------------------
string BF_StringToUpper (const string sString)
{

	char cChar;
	string sUpper = "";
	unsigned int iIndex = 0;

	while (iIndex < sString.length())
	{
		cChar = sString [iIndex];

		// if cChar between "a-z", transform into "A-Z"
		if (cChar >= 'a' && cChar <= 'z')
			cChar = cChar - 32;
		
		sUpper = sUpper + cChar;

		iIndex++;
	}

	return sUpper;

}	




