//************************************************************************
//*                          BipFILE Win V0.9.1                          *
//*----------------------------------------------------------------------*
//*                      Header file for BipFile Win                     *
//*                                                                      *
//*                                                                      *
//* Licence : NFM                                                        *
//*                                                         P. Basset    *
//************************************************************************


// include BipFILE only one time
#ifndef _BipFILE_
#define _BipFILE_

#include <fstream>
#include <sql.h>
#include <sqlext.h>
#include "BipDATE.h"
#include "BipNUM.h"


	using namespace std;



	// other define
	const string BF_BDB_BEGDESC			= "<BDBDESC>";	// BDB beginning of description
	const string BF_BDB_ENDDESC			= "</BDBDESC>";	// BDB end of description
	const string BF_BDB_BEGRECORD		= "<BDBRC>";	// BDB beginning of record
	const string BF_BDB_ENDRECORD		= "</BDBRC>";	// BDB end of record
	const char BF_BDB_BEGFIELD			= '<';			// BDB beginning of field
	const char BF_BDB_ENDFIELD			= '>';			// BDB end of field
	const char BF_BDB_DESCSEPARATOR		= ':';			// BDB separator between name and type

	const int BF_MAX_FIELD				= 20;			// max number of field
	const int BF_MAX_FIELDSIZE			= 256;			// length of char[] field

	const string BF_DB_BDB				= "BDB";		// BDB database
	const string BF_DB_ACCESS			= "ACCESS";		// ACCESS database
	const string BF_DB_SQLSERVER		= "SQLSERVER";	// SQL Server database
	const string BF_DB_MYSQL			= "MYSQL351";	// MySQL Server database v3.51

	const string BF_BIP_SQLTABLE		= "BIPressource";	// table containing BIP RSC



	// struct define
	struct BF_STRUCT_FIELD						// Field structure
	{
		string sInterFieldType;					// field type (string,int,...)
		string sDbFieldType;					// database type
		int iDbFieldOrder;						// database order number
		int iDbFieldLength;						// database field length
		bool bDbFieldNull;						// database null allowed
		string sDbFieldRemark;					// database remark
		string sDbFieldDefValue;				// database default value
		int iDbFieldDecimal;					// num of char after decimal point
		bool bIsNull;							// the value of the field is null
		string sValue;							// current value
		long lValue;							// current value
		LONGLONG llValue;						// current value
		double dValue;							// current value
		BP_Date dtValue;						// current value
		BP_Currency curValue;					// current value
	};

	struct BF_STRUCT_TABLE						// Table structure
	{
		string sDbName;							// os name of the database
		string sDbTableName;					// DB name of the table
		string sDbType;							// type of the database
		string sDbServer;						// name of the server
		SQLHENV SQLEnvir;						// environnement handle
		SQLHDBC SQLDbc;							// connection handle
		SQLHSTMT SQLStmt;						// statement handle (reading)
		SQLHSTMT SQLRecordStmt;					// handle statement (recording)
		int iStream;							// index of stream use (-1=none)
	};

	struct BF_STRUCT_STREAM						// Stream structure
	{
		bool BFInUse;							// stream is already used
		string BFMode;							// opening mode
		fstream BFStream;						// stream
	};

	struct BF_STRUCT_ERROR						// Error structure
	{
		string sName;							// internal file name
		string sFunction;						// function in use
		string sParam;							// parameters
	};



	// misc error define
	const int BF_ERR_NOTFOUND			= 1;		// file doesn't exists
	const int BF_ERR_NAME				= 2;		// bad internal file name
	const int BF_ERR_BADTYPE			= 3;		// DataBase Type unknown
	const int BF_ERR_ALROPEN			= 4;		// file is already open
	const int BF_ERR_NOTOPEN			= 5;		// file is not open
	const int BF_ERR_STREAMFULL			= 6;		// stream structure is full
	const int BF_ERR_OPEN				= 7;		// error during opening file
	const int BF_ERR_QUERYNULL			= 8;		// SQL Query is empty
	const int BF_ERR_BADFIELDNAME		= 9;		// field name is bad
	const int BF_ERR_BADFIELDTYPE		= 10;		// field type is unknown
	const int BF_ERR_BADFIELDLENGTH		= 11;		// bad field length
	const int BF_ERR_FIELDNOTFOUND		= 12;		// field doesn't exist
	const int BF_ERR_FIELDNOTSTRING		= 13;		// field is not a string
	const int BF_ERR_FIELDNOTINT		= 14;		// field is not an integer
	const int BF_ERR_FIELDNOTDATE		= 15;		// field is not a datetime
	const int BF_ERR_FIELDNOTCUR		= 16;		// field is not currency
	const int BF_ERR_BADDBTNAME			= 17;		// database table name is bad
	const int BF_ERR_FIELDEXIST			= 18;		// field already exist
	const int BF_ERR_NOPOINT			= 19;		// point char is forbidden
	const int BF_ERR_BADMODE			= 20;		// bad open mode
	const int BF_ERR_BADCOMMAND			= 21;		// SQL Command is bad or empty
	const int BF_ERR_NOPREPARE			= 22;		// use BF_RecordPrepare before



	// sql-odbc error define

	// sql-odbc error ENVIRONNEMENT define
	const int BF_ERR_SQLALLOCENV		= 1100;		// ODBC SQLAllocHandle envir
	const int BF_ERR_SQLSETENVATTR		= 1101;		// ODBC SQLSetEnvAttr

	// sql-odbc error CONNECTION define
	const int BF_ERR_SQLALLOCCONN		= 1200;		// ODBC SQLAllocHandle connection
	const int BF_ERR_SQLSETCONNATTR		= 1201;		// ODBC SQLSetConnectAttr
	const int BF_ERR_SQLDRIVERCONNECT	= 1202;		// ODBC SQLDriverConnect

	// sql-odbc error STATEMENT define
	const int BF_ERR_SQLALLOCSTATEMENT	= 1300;		// ODBC SQLAllocHandle statement
	const int BF_ERR_SQLEXECDIRECT		= 1301;		// ODBC SQLExecDirect
	const int BF_ERR_SQLGETDATA			= 1302;		// ODBC SQLGetData
	const int BF_ERR_SQLFETCHSCROLL		= 1303;		// ODBC SQLFetchScroll
	const int BF_ERR_SQLGETDESCFIELD	= 1304;		// ODBC SQLGetDescField
	const int BF_ERR_SQLCOLUMNS			= 1305;		// ODBC SQLColumns
	const int BF_ERR_SQLCLOSECURSOR		= 1306;		// ODBC SQLCloseCursor
	const int BF_ERR_SQLBINDPARAMETER	= 1307;		// ODBC SQLSQLBindParameter
	const int BF_ERR_SQLEXECUTE			= 1308;		// ODBC SQLExecute
	const int BF_ERR_SQLPREPARE			= 1309;		// ODBC SQLPrepare

	// BDB error define
	const int BF_ERR_BDBNORECORD		= 2000;		// BDB no record found
	const int BF_ERR_DESCNOTFOUND		= 2001;		// BDB Description not found
	const int BF_ERR_EDESCNOTFOUND		= 2002;		// BDB EndDescription not found
	const int BF_ERR_DESCKO				= 2003;		// BDB error in description record
	const int BF_ERR_DESCNOENDFIELD		= 2004;		// BDB no end-field found
	const int BF_ERR_DESCNONSEPAR		= 2005;		// BDB no descriptor-separator found

	// type of fields in BipFile functions
	const string BF_FLD_STRING			= "string";
	const string BF_FLD_BYTE			= "byte";
	const string BF_FLD_INTEGER			= "integer";
	const string BF_FLD_LONG			= "long";
	const string BF_FLD_LONGLONG		= "longlong";
	const string BF_FLD_FLOAT			= "float";
	const string BF_FLD_DATETIME		= "datetime";
	const string BF_FLD_CURRENCY		= "currency";
	const string BF_FLD_BINARY			= "binary";

	// string for parameters in BipFile functions
	const string BF_PRM_MODE			= "MODE";	// openning mode of stream
	const string BF_PRM_TRUST			= "TRUST";	// trusted connection
	const string BF_PRM_USER			= "USER";	// user name
	const string BF_PRM_PSW				= "PSW";	// password
	const string BF_PRM_PORT			= "PORT";	// port number
	const string BF_PRM_QUERY			= "QUERY";	// query



// Forward declarations of functions included in this code module:

	// table functions
	int BF_TableOpen (const string sBFName, const string sBFTableName, const string sBFDbName,
				  const string sBFSrvName, const string sBFDbType, const string sBFParam = "");
				  // sBFParam = ""
	int BF_TableClose (const string sBFName);
	BF_STRUCT_TABLE BF_TableGet (const string sBFTable);
	BF_STRUCT_TABLE BF_TableNullGet (void);
	int BF_SQLRecordDescriptionLoad (const string sBFName);
	int BF_SQLQueryDescriptionLoad (const string sBFName);
	int BF_BDBRecordDescriptionLoad (const string sBFName);
	int BF_BDBRecordDescriptionSave (const string sBFName, const string sBFRecord);


	// field functions
	int BF_FieldAdd (const string sBFName, const string sBFFieldName,
					const string sBFInterType, const int iBFFieldLength,
					const string sBFDbType, const bool bBFDbNull,
					const string sBFDbRemark = "", const string sBFDbDefValue = "",
					const int iBFDbDecimal = 0);
					// sBFDbRemark = "", sBFDbDefValue = "", iBFDbDecimal = 0
	BF_STRUCT_FIELD BF_FieldGet (const string sBFTable, const string sBFField);
	string BF_FieldNameGet (const string sBFTable, const int iBFOrder);
	string BF_FieldValueGetString (const string sBFName, const string sBFField);
	string BF_FieldValueGetString (const string sBFName, const int iBFOrder);
	long BF_FieldValueGetInteger (const string sBFName, const string sBFField);
	long BF_FieldValueGetInteger (const string sBFName, const int iBFOrder);
	LONGLONG BF_FieldValueGetLongLong (const string sBFName, const string sBFField);
	LONGLONG BF_FieldValueGetLongLong (const string sBFName, const int iBFOrder);
	double BF_FieldValueGetFloat (const string sBFName, const string sBFField);
	double BF_FieldValueGetFloat (const string sBFName, const int iBFOrder);
	BP_Date BF_FieldValueGetDate (const string sBFName, const string sBFField);
	BP_Date BF_FieldValueGetDate (const string sBFName, const int iBFOrder);
	BP_Currency BF_FieldValueGetCurrency (const string sBFName, const string sBFField);
	BP_Currency BF_FieldValueGetCurrency (const string sBFName, const int iBFOrder);
	BF_STRUCT_FIELD BF_FieldNullGet (void);
	int BF_FieldSetString (const string sBFName, const string sBFField,
						const string sBFValue, const bool bBFNull = false);
						// const bool bBFNull = false
	int BF_FieldSetInteger (const string sBFName, const string sBFField,
						const long lBFValue, const bool bBFNull = false);
						// const bool bBFNull = false
	int BF_FieldSetLongLong (const string sBFName, const string sBFField,
						const LONGLONG llBFValue, const bool bBFNull = false);
						// const bool bBFNull = false
	int BF_FieldSetFloat (const string sBFName, const string sBFField,
						const double dBFValue, const bool bBFNull = false);
						// const bool bBFNull = false
	int BF_FieldSetDate (const string sBFName, const string sBFField,
						BP_Date dtBFValue, const bool bBFNull = false);
						// const bool bBFNull = false
	int BF_FieldSetCurrency (const string sBFName, const string sBFField,
						BP_Currency curBFValue, const bool bBFNull = false);
						// const bool bBFNull = false
	BF_STRUCT_FIELD BF_SQLQueryFieldGet (const string sBFTable, const int iBFOrder);


	// record functions
	int BF_RecordReadNext (const string sBFName);
	int BF_RecordSave (const string sBFName);
	int BF_RecordPrepare (string sBFName);


	// SQL functions
	string BF_SQLQueryOpen (const string sBFQueryName, const string sBFSQLQuery,
					 const string sBFDbName, const string sBFSrvName,
					 const string sBFDbType, const string sBFParam = "");
				  // sBFParam = ""
	string BF_SQLCommand (const string sBFSQLCommand, const string sBFDbName,
					  const string sBFSrvName, const string sBFDbType, const string sBFParam);
	string BF_SQLConnectStringGet (const string sBFDbName, const string sBFSrvName,
							   const string sBFDbType, const string sBFParam);

	// other functions
	void BF_ResetBipFILEIntl (void);
	void BF_ErrorSet (const string sBFName, const string sBFFunction, const string sBFParam);
	string BF_ErrorDisplay (const int iBFError);
	void BF_ErrorDisplaySet (const bool bBFError = true, const bool bBFWrite = false);
				// bBFError = true, bBFWrite = false


	string BF_IntToString (const long iLong);
	int BF_StringToInt (const string sString, const int iBase = 10);
				// iBase = 10
	string BF_LongLongToString (const LONGLONG iLong);
	LONGLONG BF_StringToLongLong (const string sString);
	string BF_DoubleToString (const double dDouble);
	string BF_StringToUpper (const string sString);
	string BF_StringParamGetIntl (const string sBFParam, const string sBFString);
	bool BF_StringIsTrue (const string sBFInput);
	bool BF_StringIsFalse (const string sBFInput);

#endif /* _BipFILE_ */
